###############################################################################
#                   DMG GAUSSIAN/PROBIT MODEL SAMPLING                        #
###############################################################################

# function dmgbayes.gaussiansample
#
# Description: given data, a Gaussian directed mixed graph model, and a prior
#              for the parameters, sample from the posterior distribution of its
#              parameters. Results are stored in a text file.
#
# Input:
#
# Assume we have m directed edges, m' bi-directed edges and p variables.
#
# - var.names: array of string naming all variables (observed and latents)
# - directed: a m x 2 array, where each row (i, j) encodes a directed edge from
#             the jth variable into the ith variable. All variables should be
#             encoded as integers (1, 2, 3, ...), and their number should match
#             the corresponding entry in var.names.
# - bidirected: a m' x 2 array, m' being the total number of bi-directed edges.
#               As above, each two-dimensional row (i, j) encodes a bi-directed
#               edge connecting the ith and jth variables.
# - train: data which we condition on when computing the posterior.
# - islatent: binary array, indicating with '1' which variables in var.names are
#             unobserved variables
# - prior.V: a p x p matrix with the matrix parameter of the G-Inverse Wishart
#            prior for the covariance matrix of the error terms.
# - dmg.df: degrees of freedom of the same prior.
# - b.priormean: a (m + p) dimensional array contaning the mean parameter of the
#                Gaussian prior for each coefficient associated with the
#                respective directed edge. *IMPORTANT*: remember that we have
#                to consider the intercept terms too. The intercept term
#                for each variable Y_i should be located just after the other
#                coefficients for Y_i. This explains why this afrray has
#                m + p dimensions.
# - b.priorvar: the respective variances.
# - b.fixed: a (m + p) dimensional binary array indicating with '1' those coefficients
#            that are to be fixed to a constant value. The constant to be used
#            is the respective entry in b.priormean. *REMEMBER*: intercepts
#            should be considered.
# - mcmc.num.samples: total number of samples to be generated in the Markov chain.
# - mcmc.burn.in: total number of initial samples to be skipped when storing the
#                 results of the chain (minimum value: 0)
# - mcmc.steps: interval between points that will be stored in the output file.
# - output.filename: name of the file to which samples points will be stored.
# - mcmc.options: a numeric array contain miscellaneous options.
#
#                 mcmc.options[1]: a number between 1 and 4 indicating verbosity
#                                  level of output messages returned to the
#                                  R console.
#                 mcmc.options[2]: set to '1' if you want to generate a file
#                                  with sampled latents. This file will be called
#                                  'output.filename'.latents
# - test: optional matrix with test data to be evaluated by predictive
#         log-likelihood.
#
# It returns a 2-dimensional array. The first number is 0 if the algorithm
# for some reason halted unsuccessfully (1 if success). The second number is
# the estimated predictive log-likelihood of the model in a test set,
# if given.

dmgbayes.gaussiansample <- function(var.names, directed, bidirected, train, islatent,
                            prior.V, dmg.df, b.priormean, b.priorvar, b.fixed,
                            mcmc.num.samples, mcmc.burn.in, mcmc.step,
                            output.filename, mcmc.options, test = NULL)
{
   num.vars <- max(max(directed), max(bidirected))
   hjw <- .jnew("rbas/dmg/app/DmgExternal", .jengine(TRUE))
   dmg.var.names <- format.vector(var.names)
   dmg.directed <- format.directed.matrix(num.vars, directed)
   dmg.bidirected <- format.directed.matrix(num.vars, bidirected)
   dmg.islatent <- format.vector(as.integer(islatent))
   sample.size <- dim(train)[1]
   dmg.train <- format.matrix(train)
   dmg.prior.V <- format.matrix(prior.V)
   dmg.b.priormean <- format.vector(b.priormean)
   dmg.b.priorvar <- format.vector(b.priorvar)
   dmg.b.fixed <- format.vector(as.integer(b.fixed))
   dmg.options <- format.vector(as.integer(mcmc.options))
   if (!is.null(test)) {
       dmg.test <- format.matrix(test)
       use.test <- 1
   }
   else {
       dmg.test <- dmg.train
       use.test <- 0
   }
   out <- .jcall(hjw, "[D", "gibbsGaussianDmg", as.integer(num.vars), dmg.var.names,
                 dmg.directed, dmg.bidirected, dmg.islatent,
                 sample.size, dmg.train, dmg.test,
                 dmg.prior.V, dmg.df, dmg.b.priormean, dmg.b.priorvar, dmg.b.fixed,
                 as.integer(mcmc.num.samples), as.integer(mcmc.burn.in), as.integer(mcmc.step),
                 output.filename, dmg.options, as.integer(use.test))
   return(out)
}

# function dmgbayes.gaussian.marginal.bidirected
#
# Description: given data, a Gaussian bidirected graph model, and a prior
#              for the parameters, calculates the marginal log-likelihood of
#              the model. The model assumes means are equal to zero.
#              No latent variables are allowed.
#
# Input:
#
# Assume we have m' bi-directed edges and p variables.
#
# - var.names: array of string naming all variables (observed and latents)
# - bidirected: a m' x 2 array, m' being the total number of bi-directed edges.
#               As above, each two-dimensional row (i, j) encodes a bi-directed
#               edge connecting the ith and jth variables.
# - train: data which we condition on when computing the marginal log-likelihood.
# - prior.V: a p x p matrix with the matrix parameter of the G-Inverse Wishart
#            prior for the covariance matrix of the error terms.
# - dmg.df: degrees of freedom of the same prior.
# - mc.num.samples: total number of samples to be used in the computation.
# - mc.options: a numeric array contain miscellaneous options.
#
#                 mcmc.options[1]: a number between 1 and 4 indicating verbosity
#                                  level of output messages returned to the
#                                  R console.

dmgbayes.gaussian.marginal.bidirected <- function(var.names, bidirected, train,
                            prior.V, dmg.df, mc.num.samples, mc.options)
{
   num.vars <- max(bidirected)
   hjw <- .jnew("rbas/dmg/app/DmgExternal", .jengine(TRUE))
   dmg.var.names <- format.vector(var.names)
   dmg.bidirected <- format.directed.matrix(num.vars, bidirected)
   sample.size <- dim(train)[1]
   dmg.train <- format.matrix(train)
   dmg.prior.V <- format.matrix(prior.V)
   dmg.options <- format.vector(as.integer(mc.options))
   out <- .jcall(hjw, "[D", "marginalLLikelihoodGaussianBidirected", as.integer(num.vars), dmg.var.names,
                 dmg.bidirected, sample.size, dmg.train,
                 dmg.prior.V, dmg.df, as.integer(mc.num.samples), dmg.options)
   return(out)
}

############################################################ AUXILIARY FUNCTIONS

format.directed.matrix <- function(num.vars, directed) {
   num.edges <- dim(directed)[1];
   output <- rep(0, num.vars * num.vars)
   for (i in 1:num.edges) {
       output[1 + (directed[i, 1]  - 1) * num.vars + (directed[i, 2] - 1)] <- 1
   }
   return(.jarray(as.integer(output)))
}

format.bidirected.matrix <- function(num.vars, bidirected) {
   num.edges <- dim(bidirected)[1];
   output <- rep(0, num.vars * num.vars)
   for (i in 1:num.edges) {
       output[1 + (bidirected[i, 1] - 1) * num.vars + (bidirected[i, 2] - 1)] <- 1
       output[1 + (bidirected[i, 2] - 1) * num.vars + (bidirected[i, 1] - 1)] <- 1
   }
   return(.jarray(as.integer(output)))
}

format.matrix <- function(x) {
   stacked.data <- c()
   x.matrix <- as.matrix(x)
   for (i in 1:nrow(x.matrix))
      stacked.data <- c(stacked.data, x.matrix[i, ])
   stacked.data <- as.vector(stacked.data)
   return(.jarray(stacked.data));
}

format.vector <- function(x) {
   return(.jarray(x));
}
