package rbas.dmg.app;

import edu.cmu.tetrad.graph.*;
import edu.cmu.tetrad.data.RectangularDataSet;
import edu.cmu.tetrad.data.ContinuousVariable;
import edu.cmu.tetrad.data.ColtDataSet;

import java.util.List;
import java.util.ArrayList;
import java.io.PrintStream;

import org.rosuda.JRI.Rengine;
import org.rosuda.JRI.RConsoleOutputStream;
import rbas.dmg.sem.McmcGaussianMixedGraph;

/**
 * This class handles all the functions that are accessed by an external
 * environment. Currently it supports only the R package 'dmgbayes'.
 * This companion R package provides an interface for R applications.
 */

public class DmgExternal {

    private Rengine rEngine;

    /**
     * This function starts the Gibbs sampler from R and stores the results in a
     * external text file. 
     *
     * Arguments: please see the respective R file 'dmgsample.R'
     *
     * Returns an array 'result'. result[0] == 1 if no error happened, 0 otherwise.
     * result[1] is the predictive log-likelihood of the model if a test set has been provided.
     */

    public double[] gibbsGaussianDmg(int numVars, String varNames[], int directed[], int bidirected[], int isLatent[],
                        int sampleSize, double train[], double test[],
                        double vPriorIn[], double dfIn, double bPriorMeanIn[], double bPriorVarIn[], int fixedBIn[],
                        int mcmcSampleSize, int mcmcBurnIn, int mcmcStep, String outputFileName,
                        int options[], int useTest)
    {
        double result[] = {1.0, 0.0};
        PrintStream pR = new PrintStream(new RConsoleOutputStream(this.rEngine, 0));
        try {
            //Load graph and data
            Graph              graph = assembleDmg(numVars, varNames, directed, bidirected, isLatent);
            RectangularDataSet data = assembleData(numVars, varNames, isLatent, sampleSize, train);
            RectangularDataSet testData = null;
            if (test != null && useTest > 0)
                testData = assembleData(numVars, varNames, isLatent, sampleSize, test);

            //Load priors
            double  vPrior[][] = assembleVPrior(numVars, vPriorIn);
            double  df = dfIn;
            double  bPrior[][][] = assembleBPrior(numVars, graph, bPriorMeanIn, bPriorVarIn);
            boolean fixedB[][] = assembleFixedB(numVars, graph, fixedBIn);

            //Check options
            if (options.length != 2)
                throw new RuntimeException("Invalid number of options: " + options.length);
            if (options[0] < 1 || options[0] > 4)
                throw new RuntimeException("Invalid level of verbosity: " + options[0] +
                                           ". Choices = {1 = OFF, 2 = FINE, 3 = FINER, 4 = FINEST}.");

            //Do the sampling
            McmcGaussianMixedGraph sampler = new McmcGaussianMixedGraph(graph, data, vPrior, df, bPrior, fixedB);
            if (this.rEngine != null)
                McmcGaussianMixedGraph.resetLoggerHandler(new ExtHandler(this.rEngine));
            else
                McmcGaussianMixedGraph.resetLoggerHandler(new ExtHandler());
            McmcGaussianMixedGraph.setLoggerLevel(options[0]);
            if (options[1] == 1) {
                sampler.setReturnLatents(true);
                McmcGaussianMixedGraph.logger.finer("Storing sampled latents...");
            }

            RunGibbsGaussian.setEdgeFormat(RunGibbsGaussian.EDGE_FORMAT_R);
            result[1] = RunGibbsGaussian.inference(sampler, mcmcSampleSize, mcmcBurnIn, mcmcStep, outputFileName,
                        testData, null);
        } catch (RuntimeException e) {
            pR.println("ERROR! " + e.getMessage());
            result[0] = 0.0;
            return result;
        }
        return result;
    }


    /**
     * This function starts the Gibbs sampler from R and stores the results in a
     * external text file.
     *
     * Arguments: please see the respective R file 'dmgsample.R'
     *
     * Returns an array 'result'. result[0] == 1 if no error happened, 0 otherwise.
     * result[1] is the predictive log-likelihood of the model if a test set has been provided.
     */

    public double[] marginalLLikelihoodGaussianBidirected(int numVars, String varNames[], int bidirected[],
                        int sampleSize, double train[], double vPriorIn[], double dfIn,
                        int importanceSampleSize, int options[])
    {
        double result[] = {1.0, 0.0};
        PrintStream pR = new PrintStream(new RConsoleOutputStream(this.rEngine, 0));
        try {
            //Set some defaults
            double bPriorMeanIn[] = new double[numVars];
            double bPriorVarIn[] = new double[numVars];
            int    fixedBIn[] = new int[numVars];
            int    isLatent[] = new int[numVars];
            for (int i = 0; i < numVars; i++) {
                isLatent[i] = 0;
                bPriorMeanIn[i] = 0; bPriorVarIn[i] = 1;
                fixedBIn[i] = 1;
            }
            int directed[] = new int[numVars * numVars];
            for (int i = 0; i < numVars * numVars; i++)
                directed[i] = 0;

            //Load graph and data
            Graph              graph = assembleDmg(numVars, varNames, directed, bidirected, isLatent);
            RectangularDataSet data = assembleData(numVars, varNames, isLatent, sampleSize, train);

            //Load priors
            double  vPrior[][] = assembleVPrior(numVars, vPriorIn);
            double  df = dfIn;
            double  bPrior[][][] = assembleBPrior(numVars, graph, bPriorMeanIn, bPriorVarIn);
            boolean fixedB[][] = assembleFixedB(numVars, graph, fixedBIn);

            //Check options
            if (options.length != 1)
                throw new RuntimeException("Invalid number of options: " + options.length);
            if (options[0] < 1 || options[0] > 4)
                throw new RuntimeException("Invalid level of verbosity: " + options[0] +
                                           ". Choices = {1 = OFF, 2 = FINE, 3 = FINER, 4 = FINEST}.");

            //Do the sampling
            McmcGaussianMixedGraph sampler = new McmcGaussianMixedGraph(graph, data, vPrior, df, bPrior, fixedB);
            if (this.rEngine != null)
                McmcGaussianMixedGraph.resetLoggerHandler(new ExtHandler(this.rEngine));
            else
                McmcGaussianMixedGraph.resetLoggerHandler(new ExtHandler());
            McmcGaussianMixedGraph.setLoggerLevel(options[0]);

            result[1] = RunGaussianBidirectedScoring.scoreBidirected(sampler, importanceSampleSize);
        } catch (RuntimeException e) {
            pR.println("ERROR! " + e.getMessage());
            result[0] = 0.0;
            return result;
        }
        return result;
    }

    public static Graph assembleDmg(int numVars, String dataVar[], int directed[], int bidirected[], int isLatent[]) {
        Graph output = new EdgeListGraph();

        List<Node> variables = new ArrayList<Node>();
        for (int v = 0; v < numVars; v++) {
             Node node = new GraphNode(dataVar[v]);
             variables.add(node);
             output.addNode(node);
             if (isLatent[v] > 0)
                 node.setNodeType(NodeType.LATENT);
        }

        for (int v1 = 0; v1 < numVars; v1++) {
            Node node1 = variables.get(v1);
            for (int v2 = v1 + 1; v2 < numVars; v2++) {
                Node node2 = variables.get(v2);
                if (directed[v1 * numVars + v2] > 0)
                    output.addDirectedEdge(node2, node1);
                if (directed[v2 * numVars + v1] > 0)
                    output.addDirectedEdge(node1, node2);
                if (bidirected[v1 * numVars + v2] > 0)
                    output.addBidirectedEdge(node1, node2);
            }
        }

        return output;
    }

    /**
     * Another reminder: the data in 'dataD' should be stored *rowwise* (i.e., all variables for data point 1 come
     * before all variables for data point 2, etc.)
     */

    public static RectangularDataSet assembleData(int numVars, String varNames[], int isLatent[],
                                                  int sampleSize, double dataD[]){
        int numObserved = 0;
        for (int i = 0; i < numVars; i++)
            if (isLatent[i] == 0)
                numObserved++;
        if (dataD.length == 0)
            throw new RuntimeException("No data points in the given data!");
        if (dataD.length != numObserved * sampleSize)
            throw new RuntimeException("Size of the dataset does not correspond to the number of observed variables!");

        String columnNames[] = new String[numObserved];
        int    countN = 0;
        for (int i = 0; i < numVars; i++)
            if (isLatent[i] == 0)
                columnNames[countN++] = varNames[i];
        List<Node> columnNodes = new ArrayList<Node>();
        for (int i = 0; i < columnNames.length; i++)
             columnNodes.add(new ContinuousVariable(columnNames[i]));
        RectangularDataSet output = new ColtDataSet(sampleSize, columnNodes);
        for (int i = 0; i < sampleSize; i++) {
            for (int j = 0; j < numObserved; j++)
                 output.setDouble(i, j, dataD[i * numObserved + j]);
        }
        return output;
    }

    public static double[][] assembleVPrior(int numVars, double vPriorIn[]) {
        if (vPriorIn.length != numVars * numVars)
           throw new RuntimeException("Size of covariance hyperparameter does not correspond to the number of variables!");
        double vPrior[][] = new double[numVars][numVars];
        for (int v1 = 0; v1 < numVars; v1++)
            for (int v2 = 0; v2 < numVars; v2++)
                vPrior[v1][v2] = vPriorIn[v1 * numVars + v2];
        return vPrior;
    }

    public static double[][][] assembleBPrior(int numVars, Graph graph, double bPriorMeanIn[], double bPriorVarIn[]) {
        double bPrior[][][] = new double[numVars][][];
        int    countIn = 0;
        for (int i = 0; i < numVars; i++) {
            Node node = (Node) graph.getNodes().get(i);
            bPrior[i] = new double[graph.getParents(node).size() + 1][2];
            for (int j = 0; j < graph.getParents(node).size() + 1; j++) {
                if (countIn >= bPriorMeanIn.length || countIn >= bPriorVarIn.length)
                   throw new RuntimeException("Number of coefficient hyperparameters does not " +
                                              "correspond to the number of coefficients!");
                bPrior[i][j][0] = bPriorMeanIn[countIn];
                if (bPriorVarIn[countIn] <= 0)
                    throw new RuntimeException("Invalid prior variance for coefficients into variable " + (i + 1));
                bPrior[i][j][1] = bPriorVarIn[countIn];
                countIn++;
            }
        }
        return bPrior;
    }

    public static boolean[][] assembleFixedB(int numVars, Graph graph, int fixedBIn[]) {
        boolean fixedB[][] = new boolean[numVars][];
        int countIn = 0;
        for (int i = 0; i < numVars; i++) {
            Node node = (Node) graph.getNodes().get(i);
            fixedB[i] = new boolean[graph.getParents(node).size() + 1];
            for (int j = 0; j < graph.getParents(node).size() + 1; j++) {
                if (countIn >= fixedBIn.length)
                   throw new RuntimeException("Number of coefficient fixed indicators does not " +
                                              "correspond to the number of coefficients!");
                fixedB[i][j] = (fixedBIn[countIn] == 1);
                countIn++;
            }
        }
        return fixedB;
    }

    public DmgExternal() {
        this.rEngine = null;
    }

    public DmgExternal(Rengine rEngine) {
        this.rEngine = rEngine;
    }

    public static void main(String[] args) {
    }
}

