package rbas.dmg.app;

import edu.cmu.tetrad.graph.*;
import edu.cmu.tetrad.data.RectangularDataSet;
import edu.cmu.tetrad.data.ContinuousVariable;
import edu.cmu.tetrad.data.ColtDataSet;

import java.util.*;
import java.util.logging.Logger;
import java.util.logging.Level;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.PrintWriter;

/**
 * Loader
 *
 * Implements methods for loading different types of data structures from text files
 * (i.e., graphs, priors and datasets)
 *
 * @author Ricardo Silva
 * @version $Revision: 1.0 $ $Date: 2007/07/02$
 */

public class Loader {

    public static Logger logger = Logger.getLogger("rbas.dmg.app.Loader");

    public static void startLogger() {
        Loader.logger.setLevel(Level.FINE);
        Loader.logger.addHandler(new ExtHandler());
    }

    public static Graph loadGraph(String fileName, List<String> dataVariablesList) throws java.io.IOException {
        Graph output = new EdgeListGraph();
        BufferedReader in = new BufferedReader(new FileReader(fileName));
        Hashtable<String, Node> variables = new Hashtable<String, Node>();
        HashSet<String> observedVariables = new HashSet<String>();
        String dataVar[] = new String[dataVariablesList.size()];
        for (int v = 0; v < dataVar.length; v++) {
            dataVar[v] = dataVariablesList.get(v);
            observedVariables.add(dataVar[v]);
            Node node = new GraphNode(dataVar[v]);
            variables.put(dataVar[v], node);
            output.addNode(node);
        }
        int lineCount = 0;

        while (true) {
            String line = in.readLine();
            if (line == null)
                break;
            line = line.trim();
            lineCount++;
            if (line.equals(""))
                continue;

            String tokens[] = line.split(" +|\t");
            if (tokens.length != 3 || !(tokens[1].equals("->") || tokens[1].equals("<->")))
                throw new RuntimeException("Cannot parse file " + fileName + ". Syntax error! Line " + lineCount + ": " + line);
            Node node1 = variables.get(tokens[0]), node2 = variables.get(tokens[2]);
            if (node1 == null) {
                node1 = new GraphNode(tokens[0]);
                variables.put(tokens[0], node1);
                output.addNode(node1);
                if (!observedVariables.contains(tokens[0]) && observedVariables.size() > 0) {
                    node1.setNodeType(NodeType.LATENT);
                    Loader.logger.fine("Setting variable " + node1.getName() + " to latent");
                }
            }
            if (node2 == null) {
                node2 = new GraphNode(tokens[2]);
                variables.put(tokens[2], node2);
                output.addNode(node2);
                if (!observedVariables.contains(tokens[2]) && observedVariables.size() > 0) {
                    node2.setNodeType(NodeType.LATENT);
                    Loader.logger.fine("Setting variable " + node2.getName() + " to latent");
                }
            }
            if (tokens[1].equals("->"))
                output.addDirectedEdge(node1, node2);
            else
                output.addBidirectedEdge(node1, node2);
        }
        return output;
    }

    public static RectangularDataSet loadDataSet(String fileName) throws java.io.IOException {
        BufferedReader in = new BufferedReader(new FileReader(fileName));
        int lineCount = 0;

        String line;
        do {
            line = in.readLine();
            if (line != null)
               line = line.trim();
            lineCount++;
        } while (line != null && line.equals(""));

        if (line == null)
            throw new RuntimeException("Invalid dataset: " + fileName + "!");

        String columnNames[] = line.split(" +|\t+"); //header
        List<List<Double>> data = new ArrayList<List<Double>>();
        for (int i = 0; i < columnNames.length; i++)
             data.add(new ArrayList<Double>());
        while (true) {
            line = in.readLine();
            if (line == null)
                break;
            line = line.trim();
            lineCount++;
            if (line.equals(""))
                continue;
            String row[] = line.split(" +|\t+");
            if (row.length != columnNames.length)
                throw new RuntimeException("Number of columns in line " + lineCount +
                                           " does not match number of variables!");
            int cCount = 0;
            for (List<Double> dColumn : data)
                dColumn.add(new Double(row[cCount++]));
        }

        List<Node> columnNodes = new ArrayList<Node>();
        for (int i = 0; i < columnNames.length; i++)
             columnNodes.add(new ContinuousVariable(columnNames[i]));
        RectangularDataSet output = new ColtDataSet(data.get(0).size(), columnNodes);
        int rows = data.get(0).size(), cols = columnNames.length;
        for (int j = 0; j < cols; j++) {
            List<Double> dColumn = data.get(j);
            for (int i = 0; i < rows; i++)
                output.setDouble(i, j, (double) dColumn.get(i));
        }
        return output;
    }

    public static void loadPriors(String fileName, double vPrior[][], double df[], double bPrior[][][],
                                  boolean fixedB[][], double thresholdPrior[][][], boolean fixedT[][],
                                  Graph graph) throws java.io.IOException {
        for (int i = 0; i < vPrior.length; i++) {
            for (int j = 0; j < vPrior.length; j++)
                vPrior[i][j] = 0.;
            vPrior[i][i] = 1.;
        }
        for (int i = 0; i < graph.getNodes().size(); i++) {
            Node node = (Node) graph.getNodes().get(i);
            bPrior[i] = new double[graph.getParents(node).size() + 1][2];
            fixedB[i] = new boolean[graph.getParents(node).size() + 1];
            for (int j = 0; j < graph.getParents(node).size() + 1; j++) {
                bPrior[i][j][0] = 0.; bPrior[i][j][1] = 1.;
                fixedB[i][j] = false;
            }
        }

        BufferedReader in = new BufferedReader(new FileReader(fileName));
        int lineCount = 0;
        String line;
        do {
            line = in.readLine();
            if (line != null)
                line = line.trim();
            lineCount++;
        } while (line != null && line.equals(""));
        if (line == null)
            throw new RuntimeException("Invalid priors file: " + fileName + "!");

        String dfStr[] = line.split(" +|\t+");
        if (dfStr.length != 2 || !dfStr[0].equals("df"))
            throw new RuntimeException("Error at line " + lineCount + "! Expecting degrees of freedom");
        df[0] = Double.parseDouble(dfStr[1]);

        while (true) {
            line = in.readLine();
            if (line == null)
                break;
            line = line.trim();
            lineCount++;
            if (line.equals(""))
                continue;

            String tokens[] = packTokens(line.split(" +|\t+"));
            if (tokens.length < 4 || !(tokens[1].equals("->") || tokens[1].equals("<->") || tokens[1].equals("|")))
                throw new RuntimeException("Syntax error! Line " + lineCount + ": " + line);

            Node node1 = graph.getNode(tokens[0]), node2 = graph.getNode(tokens[2]);
            if ((node1 == null && !tokens[0].equals("_INT")) || (node2 == null && !tokens[1].equals("|")))
                throw new RuntimeException("Error! Pair (" + tokens[0] + ", " + tokens[2] + ") not in graph!");
            if (tokens[1].equals("->")) {
                if (tokens.length < 5)
                    throw new RuntimeException("Syntax error! Line " + lineCount + ": " + line);
                if (tokens[0].equals("_INT")) {
                    int n2Idx = -1;
                    for (int i = 0; i < graph.getNodes().size(); i++)
                        if (graph.getNodes().get(i) == node2) {
                            n2Idx = i;
                            break;
                        }
                    int numParents = graph.getParents(node2).size();
                    bPrior[n2Idx][numParents][0] = Double.parseDouble(tokens[3]);
                    bPrior[n2Idx][numParents][1] = Double.parseDouble(tokens[4]);
                    if (tokens.length >= 6 && tokens[5].equals("true"))
                        fixedB[n2Idx][numParents] = true;
                    continue;
                }
                List<Edge> edges = graph.getEdges(node1, node2);
                boolean found = false;
                for (Edge edge : edges) {
                    if (edge.getEndpoint1() == Endpoint.TAIL && edge.getEndpoint2() == Endpoint.ARROW) {
                        found = true;
                        break;
                    }
                }
                if (!found)
                    throw new RuntimeException("Error! Edge "+ tokens[0] + " -> " + tokens[2] + " not in graph!");
                int n2Idx = -1;
                for (int i = 0; i < graph.getNodes().size(); i++)
                    if (graph.getNodes().get(i) == node2) {
                        n2Idx = i;
                        break;
                    }
                List<Node> parents = graph.getParents(node2);
                for (int i = 0; i < parents.size(); i++)
                    if (parents.get(i) == node1) {
                        bPrior[n2Idx][i][0] = Double.parseDouble(tokens[3]);
                        bPrior[n2Idx][i][1] = Double.parseDouble(tokens[4]);
                        if (tokens.length >= 6 && tokens[5].equals("true"))
                            fixedB[n2Idx][i] = true;
                        break;
                    }
            } else if (tokens[1].equals("<->")) {
                if (tokens.length < 4)
                    throw new RuntimeException("Syntax error! Line " + lineCount + ": " + line);
                if (node1 == node2) {
                    int n1Idx = -1;
                    for (int i = 0; i < graph.getNodes().size(); i++)
                        if (graph.getNodes().get(i) == node1) {
                            n1Idx = i;
                            break;
                        }
                    vPrior[n1Idx][n1Idx] = Double.parseDouble(tokens[3]) * df[0];
                    continue;
                }
                List<Edge> edges = graph.getEdges(node1, node2);
                boolean found = false;
                for (Edge edge : edges) {
                    if (edge.getEndpoint1() == Endpoint.ARROW && edge.getEndpoint2() == Endpoint.ARROW) {
                        found = true;
                        break;
                    }
                }
                if (!found)
                    throw new RuntimeException("Error! Edge "+ tokens[0] + " <-> " + tokens[2] + " not in graph!");
                int n1Idx = -1, n2Idx = -1;
                for (int i = 0; i < graph.getNodes().size(); i++)
                    if (graph.getNodes().get(i) == node1) {
                        n1Idx = i;
                        break;
                    }
                for (int i = 0; i < graph.getNodes().size(); i++)
                    if (graph.getNodes().get(i) == node2) {
                        n2Idx = i;
                        break;
                    }
                vPrior[n1Idx][n2Idx] = vPrior[n2Idx][n1Idx] = Double.parseDouble(tokens[3]) * df[0];
            } else if (tokens[1].equals("|")) {
                int nIdx = -1;
                for (int i = 0; i < graph.getNodes().size(); i++)
                    if (graph.getNodes().get(i) == node1) {
                        nIdx = i;
                        break;
                    }
                if (thresholdPrior[nIdx] == null)
                    throw new RuntimeException("Node " + node1 + " is indicated as continuous, but was given " +
                                               "threshold priors");
                int currentTokenPoint = 2;
                for (int t = 0; t < thresholdPrior[nIdx].length; t++) {
                    if (tokens.length < currentTokenPoint + 2 + 1)
                        throw new RuntimeException("Missing information about threshold priors for node " + node1);
                    thresholdPrior[nIdx][t][0] = Double.parseDouble(tokens[currentTokenPoint]);
                    thresholdPrior[nIdx][t][1] = Double.parseDouble(tokens[currentTokenPoint + 1]);
                    currentTokenPoint += 2;
                    if (tokens.length >= currentTokenPoint + 1 && tokens[currentTokenPoint].equals("*")) {
                        fixedT[nIdx][t] = true;
                        currentTokenPoint++;
                    }
                    else
                        fixedT[nIdx][t] = false;

                    if (t < thresholdPrior[nIdx].length - 1) {
                        if (tokens.length < currentTokenPoint + 1 || !tokens[currentTokenPoint].equals("|"))
                            throw new RuntimeException("Syntax error! Line " + lineCount + ": " + line);
                        currentTokenPoint++;
                    } else if (tokens.length >= currentTokenPoint + 1)
                        throw new RuntimeException("Too many threshold points for node " + node1);
                }
            } else
                throw new RuntimeException("Syntax error! Line " + lineCount + ": " + line);
        }
    }

    public static double[][] loadTypes(String fileName, Graph graph) throws java.io.IOException {
        double output[][] = new double[graph.getNumNodes()][];

        BufferedReader in = new BufferedReader(new FileReader(fileName));
        int lineCount = 0;
        String line;

        while (true) {
            line = in.readLine();
            if (line == null)
                break;
            line = line.trim();
            lineCount++;
            if (line.equals(""))
                continue;

            String tokens[] = packTokens(line.split(" +|\t+|:+|,+"));
            if (tokens.length < 2)
                throw new RuntimeException("Syntax error! Line " + lineCount + ": " + line);

            Node node1 = graph.getNode(tokens[0]);
            if (node1 == null)
                throw new RuntimeException("Error! Node " + tokens[0] + " not in graph!");

            int nIdx = -1;
                for (int i = 0; i < graph.getNodes().size(); i++)
                    if (graph.getNodes().get(i) == node1) {
                        nIdx = i;
                        break;
                    }
            if (tokens[1].equals("continuous") || tokens[1].equals("continuous.")) {
                output[nIdx] = null;
                continue;
            }
            if (tokens.length < 3)
               throw new RuntimeException("Error! Node " + tokens[0] + " cannot have a single value!");
            List<Double> values = new ArrayList<Double>();
            for (int i = 1; i < tokens.length; i++)
                values.add(Double.parseDouble(tokens[i]));
            output[nIdx] = new double[values.size()];
            for (int i = 0; i < values.size(); i++)
                output[nIdx][i] = values.get(i);
            Arrays.sort(output[nIdx]);
        }

        return output;
    }

    protected static String[] packTokens(String tokensBase[]) {
        int totalNonNull = 0;
        for (int i = 0; i < tokensBase.length; i++)
            if (tokensBase[i].length() > 0) totalNonNull++;
        String tokens[] = new String[totalNonNull]; int pos = 0;
        for (int i = 0; i < tokensBase.length; i++)
            if (tokensBase[i].length() > 0)
                tokens[pos++] = tokensBase[i];
        return tokens;
    }

    /**
     * The simplest possible generator of cross validation files. It is given a file name that is
     * assumed to have variable names in the first line, and no gaps between lines.
     *
     * It is assumed that the original file has a ".dat" extension.
     */

    public static void generateCrossValidation(String fileName, int numFolders) {
       try {
           BufferedReader in = new BufferedReader(new FileReader(fileName + ".dat"));
           String header = in.readLine();
           List<String> lines = new ArrayList<String>();
           String line = in.readLine();
           while (line != null) {
               lines.add(line);
               line = in.readLine();
           }
           in.close();
           int numSamples = lines.size();
           Random r = new Random();
           for (int i = 0; i < 1000000; i++) {
               int l1 = r.nextInt(numSamples), l2 = r.nextInt(numSamples);
               String buffer = lines.get(l1);
               lines.set(l1, lines.get(l2));
               lines.set(l2, buffer);
           }
           int foldSize = (int) (((double) numSamples) / numFolders);
           int lastFoldSize = numSamples - foldSize * (numFolders - 1);
           int currentPoint = 0;
           for (int i = 0; i < numFolders; i++) {
               List<String> training = new ArrayList<String>();
               List<String> testing = new ArrayList<String>();
               int iSize;
               if (i < numFolders - 1)
                   iSize = foldSize;
               else
                   iSize = lastFoldSize;
               for (int j = currentPoint; j < currentPoint + iSize; j++)
                   testing.add(lines.get(j));
               for (int j = 0; j < currentPoint; j++)
                   training.add(lines.get(j));
               for (int j = currentPoint + iSize; j < lines.size(); j++)
                   training.add(lines.get(j));
               currentPoint += iSize;
               PrintWriter out1 = new PrintWriter(fileName + "." + (i + 1) + ".train");
               out1.println(header);
               for (int j = 0; j < training.size(); j++)
                   out1.println(training.get(j));
               out1.close();
               PrintWriter out2 = new PrintWriter(fileName + "." + (i + 1) + ".test");
               out2.println(header);
               for (int j = 0; j < testing.size(); j++)
                   out2.println(testing.get(j));
               out2.close();
           }

       } catch (java.io.IOException e) {
          throw new RuntimeException(e);
       }
    }
}
