package rbas.dmg.app;

import java.util.*;

import edu.cmu.tetrad.graph.*;
import edu.cmu.tetrad.data.*;
import edu.cmu.tetrad.util.MatrixUtils;

import java.io.*;

import rbas.dmg.sem.McmcGaussianMixedGraph;

/**
 * RunGibbsGaussian
 *
 * Application for generating samples from the posterior of a Gaussian mixed graph model. Optionally, it
 * also computes predictive log-likelihood in a test set.
 *
 * Currently not supporting cyclic graphs.
 *
 * @author Ricardo Silva
 * @version $Revision: 0.1 $ $Date: 2007/07/05$
 */

public class RunGibbsGaussian {

    public static int EDGE_FORMAT_GEN = 1;
    public static int EDGE_FORMAT_R = 2;

    protected static int edgeFormat = 1;

    protected static int NUM_CHAINS = 1;

    public static void main(final String[] argv) {
        System.out.println();
        System.out.println("** DMG MODELING SOFTWARE: GIBBS SAMPLING FOR GAUSSIAN MODELS");
        System.out.println("Code: Ricardo Silva, Statslab, University of Cambridge");
        System.out.println("Version: 1.0 - August 2008");

        if (argv.length != 4) {
            System.out.println("Usage: rbas.dmg.app.RunGibbsGaussian <domain_name> <#samples> <#burn in> <step size>");
            System.out.println("Requires: files domain_name.graph, domain_name.train, domain_name.priors");
            System.out.println("Optional: file domain_name.test");
            System.exit(0);
        }

        try {
            Loader.startLogger();
            System.out.println("Loading... " + argv[0] + ".train");
            RectangularDataSet data = Loader.loadDataSet(argv[0] + ".train");
            RectangularDataSet test = null;

            try {
                test = Loader.loadDataSet(argv[0] + ".test");
            } catch (java.io.IOException e) {
                System.out.println("No test set found");
            }
            Graph graph = Loader.loadGraph(argv[0] + ".graph", data.getVariableNames());
            rbas.util.RbasGraphUtils.printEdges(graph);
            double vPrior[][] = new double[graph.getNodes().size()][graph.getNodes().size()];
            double df[] = new double[1];
            double bPrior[][][] = new double[graph.getNodes().size()][][];
            boolean fixedB[][] = new boolean[graph.getNodes().size()][];
            Loader.loadPriors(argv[0] + ".priors", vPrior, df, bPrior, fixedB, null, null, graph);
            McmcGaussianMixedGraph sampler = new McmcGaussianMixedGraph(graph, data, vPrior, df[0], bPrior, fixedB);
            McmcGaussianMixedGraph.resetLoggerHandler(new ExtHandler()); McmcGaussianMixedGraph.setLoggerLevel(4);
            if (NUM_CHAINS == 1)
                inference(sampler, Integer.parseInt(argv[1]), Integer.parseInt(argv[2]), Integer.parseInt(argv[3]), argv[0], test, null);
            else if (NUM_CHAINS > 1)
                multiInference(sampler, Integer.parseInt(argv[1]), Integer.parseInt(argv[2]), Integer.parseInt(argv[3]), NUM_CHAINS, argv[0]);
            else
                throw new RuntimeException("Invalid number of chains: " + NUM_CHAINS);
        } catch (java.io.IOException e) {
            System.out.println("ERROR: " + e.getMessage());
            System.exit(0);
        } catch (RuntimeException e) {
            System.out.println("ERROR: " + e.getMessage());
            System.exit(0);
        }
	}

    /**
     * Run a single chain
     */

    public static double inference(McmcGaussianMixedGraph mcmc, int simulationSize, int burnIn, int step, String fileName,
                                   RectangularDataSet test, List samplesHere) {
        java.util.Date date1 = new java.util.Date();
        List samples = mcmc.doSample(simulationSize);
        java.util.Date date2 = new java.util.Date();
        McmcGaussianMixedGraph.logger.fine("Started at " + date1.toString() + ", finished at " + date2.toString());
        Graph graph = mcmc.getGraph();
        int numNodes = graph.getNodes().size();
        double estimate[][] = new double[numNodes][numNodes];
        for (int i = 0; i < numNodes; i++)
            for (int j = 0; j < numNodes; j++)
                estimate[i][j] = 0.;
        double iMinusB[][] = new double[numNodes][numNodes];
        int    parents[][] = rbas.util.RbasGraphUtils.getParentsMatrix(graph);
        double paramSample[] = new double[2];
        for (int i = 0; i < paramSample.length; i++)
             paramSample[i] = 0.;
        double totalWeight = 0.;
        double allBidirected[][][] = (double[][][]) samples.get(0);
        double allWeights[] = (double []) samples.get(1);
        double allDirected[][][] = (double[][][]) samples.get(2);
        for (int s = burnIn; s < simulationSize; s += step) {
            double errSample[][] = mcmc.unpackCovariance(allBidirected[s]);
            double weight = allWeights[s];
            double coeffSample[][] = allDirected[s];
            for (int i = 0; i < numNodes; i++)
                for (int j = 0; j < numNodes; j++)
                    iMinusB[i][j] = 0.;
            for (int i = 0; i < numNodes; i++)
                iMinusB[i][i] = 1.;
            for (int i = 0; i < numNodes; i++)
                for (int j = 0; j < parents[i].length; j++)
                    iMinusB[i][parents[i][j]] -= coeffSample[i][j];
            double invIMinusB[][] = MatrixUtils.inverse(iMinusB);
            double implCov[][] = MatrixUtils.product(invIMinusB,
                                     MatrixUtils.product(errSample, MatrixUtils.transpose(invIMinusB)));
            for (int i = 0; i < numNodes; i++)
                for (int j = 0; j < numNodes; j++)
                    estimate[i][j] += weight * implCov[i][j];
            totalWeight += weight;
        }
        for (int i = 0; i < numNodes; i++)
            for (int j = 0; j < numNodes; j++)
                estimate[i][j] /= totalWeight;

        McmcGaussianMixedGraph.logger.finer("-----------------------------------------------------");
        McmcGaussianMixedGraph.logger.finer("The graph:");
        McmcGaussianMixedGraph.logger.finer(graph.toString());
        McmcGaussianMixedGraph.logger.finer("");
        McmcGaussianMixedGraph.logger.finer("-----------------------------------------------------");
        McmcGaussianMixedGraph.logger.finer("Sample covariance matrix:");
        McmcGaussianMixedGraph.logger.finer("");
        RectangularDataSet data = mcmc.getReorderedData();
        double sampleM[][] = (new CovarianceMatrix(data)).getMatrix().toArray();
        String varNames = new String();
        for (int i = 0; i < numNodes; i++)
            if (graph.getNode(data.getVariable(i).getName()).getNodeType() == NodeType.MEASURED)
                varNames += data.getVariable(i).getName() + "\t";
        McmcGaussianMixedGraph.logger.finer(varNames);
        for (int i = 0; i < numNodes; i++) {
            if (graph.getNode(data.getVariable(i).getName()).getNodeType() != NodeType.MEASURED)
                continue;
            String matrixRow = new String();
            for (int j = 0; j < numNodes; j++)
                if (graph.getNode(data.getVariable(j).getName()).getNodeType() == NodeType.MEASURED)
                    matrixRow += ((int) (10000. * sampleM[i][j])) / 10000. + "\t";
            McmcGaussianMixedGraph.logger.finer(matrixRow);
        }
        McmcGaussianMixedGraph.logger.finer("");
        McmcGaussianMixedGraph.logger.finer("-----------------------------------------------------");
        McmcGaussianMixedGraph.logger.finer("Estimated matrix (expected posterior):");
        McmcGaussianMixedGraph.logger.finer("");
        for (int i = 0; i < numNodes; i++) {
            if (graph.getNode(data.getVariable(i).getName()).getNodeType() != NodeType.MEASURED)
                continue;
            String matrixRow = new String();
            for (int j = 0; j < numNodes; j++)
                if (graph.getNode(data.getVariable(j).getName()).getNodeType() == NodeType.MEASURED)
                    matrixRow += (((int) (10000. * estimate[i][j])) / 10000. + "\t");
            McmcGaussianMixedGraph.logger.finer(matrixRow);
        }
        if (samplesHere != null) {
            samplesHere.clear();
            samplesHere.addAll(samples);
        }
        try {
            McmcGaussianMixedGraph.logger.fine("Saving MCMC run...");
            dumpSampling(fileName, samples, mcmc.getGraph(), simulationSize, burnIn, step, mcmc);
            if (test != null) {
                McmcGaussianMixedGraph.logger.fine("-----------------------------------------------------");
                double ll = mcmc.predictiveLoglikelihood(samples, test, burnIn, simulationSize, step);
                McmcGaussianMixedGraph.logger.fine("Predictive loglikelihood: " + ll);
                McmcGaussianMixedGraph.logger.fine("Predictive loglikelihood per data point: " + (ll / test.getNumRows()));
                return ll / test.getNumRows();
            }
        } catch (IOException e) {
            McmcGaussianMixedGraph.logger.fine("(No test set found)");
        }
        McmcGaussianMixedGraph.logger.fine("Done!");
        return 0;
    }

    /**
     * Run several chains. Implementation is still sequential.
     */

    public static void multiInference(McmcGaussianMixedGraph mcmc, int simulationSize, int burnIn, int step,
                                      int numChains, String fileName) {
        double results[] = new double[numChains];
        List   samplesHere = new ArrayList();
        for (int i = 0; i < numChains; i++) {
            McmcGaussianMixedGraph.logger.fine(">>>>>>>>>>>>> CHAIN #" + (i + 1));
            results[i] = inference(mcmc, simulationSize, burnIn, step, fileName, null, samplesHere);
            McmcGaussianMixedGraph.logger.fine("Saving MCMC run...");
            try {
               dumpSamplingObs(fileName, samplesHere, mcmc.getGraph(), simulationSize, burnIn, step, mcmc, "." + (i + 1));
            } catch (java.io.IOException e) {
               throw new RuntimeException(e);
            }
        }
        McmcGaussianMixedGraph.logger.fine("All MCMC runs: ");
        for (int i = 0; i < numChains; i++)
            McmcGaussianMixedGraph.logger.fine(Double.toString(results[i]));
        McmcGaussianMixedGraph.logger.fine("Done!");
    }

    /**
     * Save model parameters in a file.
     */

    protected static void dumpSampling(String fileName, List samples, Graph graph, int simulationSize, int burnIn,
                                       int step, McmcGaussianMixedGraph mcmc) throws IOException {
        Graph bidirectedGraph = rbas.util.RbasGraphUtils.getBidirectedSubgraph(graph);
        int numNodes = graph.getNumNodes();

        List<String[]> directedEdges = new ArrayList<String[]>();   //edges need to reordered according to order of parameters in 'samples'
        List<String[]> bidirectedEdges = new ArrayList<String[]>(); //edges need to reordered according to order of parameters in 'samples'
        int preSpouses[][] = new int[numNodes][];

        int bidirAdj[][] = new int[numNodes][numNodes];
        for (int i = 0; i < numNodes; i++) {
            Node node1 = graph.getNodes().get(i);
            bidirAdj[i][i] = 1;
            int numPreSpouses = 0;
            for (int j = 0; j < numNodes; j++) {
                Node node2 = graph.getNodes().get(j);
                if (graph.isParentOf(node2, node1)) {
                    String newEdge[] = {node2.getName(), node1.getName()};
                    directedEdges.add(newEdge);//(node2 + " -> " + node1);
                }
                if ((j < i && bidirectedGraph.isAdjacentTo(node1, node2)) || (j == i)) {
                    String newEdge[] = {node1.getName(), node2.getName()};
                    bidirectedEdges.add(newEdge);//(node1.getName() + " <-> " + node2.getName());
                    bidirAdj[i][j] = 1;
                    numPreSpouses++;
                }
            }
            String newIntEdge[] = {"", node1.getName()}; //"Intercept-to-node" edge
            directedEdges.add(newIntEdge);//("_INT -> " + node1);
            preSpouses[i] = new int[numPreSpouses];
            int psCount = 0;
            for (int j = 0; j < i; j++)
                if (bidirAdj[i][j] > 0)
                    preSpouses[i][psCount++] = j;
            preSpouses[i][numPreSpouses - 1] = i;
        }
        PrintWriter logFile = new PrintWriter(fileName + ".mcmc");
        boolean latentSampled = (samples.size() == 4);
        PrintWriter latentFile = null;
        int numLatents = 0, sampleSize = 0;
        double allLatents[][][] = null;
        if (latentSampled) {
            latentFile = new PrintWriter(fileName + ".latents.mcmc");
            for (int i = 0; i < numNodes; i++)
                if (graph.getNodes().get(i).getNodeType() == NodeType.LATENT) {
                    latentFile.print(graph.getNodes().get(i).getName() + "\t");
                    numLatents++;
                }
            latentFile.println();
            allLatents = (double[][][]) samples.get(3);
            sampleSize = allLatents[0].length;
        }
        for (String[] edgeStr : directedEdges)
            logFile.print(formatEdge(edgeStr[0], edgeStr[1], 1) + "\t");
        for (String[] edgeStr : bidirectedEdges)
            logFile.print(formatEdge(edgeStr[0], edgeStr[1], 2) + "\t");
        logFile.println();
        double allBidirected[][][] = (double[][][]) samples.get(0);
        double allDirected[][][] = (double[][][]) samples.get(2);
        for (int s = burnIn; s < simulationSize; s += step) {
            double errSample[][] = mcmc.unpackCovariance(allBidirected[s]);
            double coeffSample[][] = allDirected[s];
            for (int i = 0; i < numNodes; i++)
                for (int p = 0; p < coeffSample[i].length; p++)
                    logFile.print(coeffSample[i][p] + "\t");
            for (int i = 0; i < numNodes; i++)
                for (int p = 0; p < preSpouses[i].length; p++)
                    logFile.print(errSample[i][preSpouses[i][p]] + "\t");
            logFile.println();
            if (latentSampled)
                for (int sl = 0; sl < sampleSize; sl++) {
                    for (int i = 0; i < numLatents; i++)
                         latentFile.print(allLatents[s][sl][i] + "\t");
                    latentFile.println();
                }
        }
        logFile.close();
        if (latentSampled)
            latentFile.close();
    }

    /**
     * Outputs the implied covariance matrix instead of the internal model parameters..
     */

    protected static void dumpSamplingObs(String fileName, List samples, Graph graph, int simulationSize, int burnIn,
                                          int step, McmcGaussianMixedGraph mcmc, String label) throws IOException {
        PrintWriter logFile = new PrintWriter(fileName + ".obs" + label + ".mcmc");

        List<String> name = new ArrayList<String>();
        for (Node node : graph.getNodes())
            if (node.getNodeType() == NodeType.MEASURED)
                name.add(node.getName());
        Object nameArray[] = name.toArray();
        java.util.Arrays.sort(nameArray);
        Hashtable<String, Integer> namePos = new Hashtable<String, Integer>();
        for (int g = 0; g < graph.getNumNodes(); g++) {
            Node node = graph.getNodes().get(g);
            if (node.getNodeType() == NodeType.LATENT)
                continue;
            for (int i = 0; i < nameArray.length; i++)
                if (nameArray[i].equals(node.getName())) {
                    namePos.put((String) nameArray[i], new Integer(g));
                    break;
                }
        }


        Graph bidirectedGraph = rbas.util.RbasGraphUtils.getBidirectedSubgraph(graph);
        int numNodes = graph.getNumNodes(), numObsNodes = name.size();

        int preSpouses[][] = new int[numNodes][];
        int bidirAdj[][] = new int[numNodes][numNodes];
        for (int i = 0; i < numNodes; i++) {
            Node node1 = graph.getNodes().get(i);
            bidirAdj[i][i] = 1;
            int numPreSpouses = 0;
            for (int j = 0; j < numNodes; j++) {
                Node node2 = graph.getNodes().get(j);
                if ((j < i && bidirectedGraph.isAdjacentTo(node1, node2)) || (j == i)) {
                    bidirAdj[i][j] = 1;
                    numPreSpouses++;
                }
            }
            preSpouses[i] = new int[numPreSpouses];
            int psCount = 0;
            for (int j = 0; j < i; j++)
                if (bidirAdj[i][j] > 0)
                    preSpouses[i][psCount++] = j;
            preSpouses[i][numPreSpouses - 1] = i;
        }
        for (int i = 0; i < numObsNodes; i++)
            for (int j = i; j < numObsNodes; j++)
                logFile.print("X" + (i + 1) + ".X" + (j + 1) + "\t");
        logFile.println();
        double allBidirected[][][] = (double[][][]) samples.get(0);
        double allDirected[][][] = (double[][][]) samples.get(2);
        double iMinusB[][] = new double[numNodes][numNodes];
        int    parents[][] = rbas.util.RbasGraphUtils.getParentsMatrix(graph);
        for (int s = burnIn; s < simulationSize; s += step) {
            double errSample[][] = mcmc.unpackCovariance(allBidirected[s]);
            double coeffSample[][] = allDirected[s];
            for (int i = 0; i < numNodes; i++)
                for (int j = 0; j < numNodes; j++)
                    iMinusB[i][j] = 0.;
            for (int i = 0; i < numNodes; i++)
                iMinusB[i][i] = 1.;
            for (int i = 0; i < numNodes; i++)
                for (int j = 0; j < parents[i].length; j++)
                    iMinusB[i][parents[i][j]] -= coeffSample[i][j];
            double invIMinusB[][] = MatrixUtils.inverse(iMinusB);
            double implCov[][] = MatrixUtils.product(invIMinusB,
                                     MatrixUtils.product(errSample, MatrixUtils.transpose(invIMinusB)));
            for (int i = 0; i < numObsNodes; i++)
                for (int j = i; j < numObsNodes; j++) {
                    double rounded = ((int) (implCov[namePos.get((String)nameArray[i])][namePos.get((String)nameArray[j])]
                                     * 1000)) / 1000.;
                    logFile.print(rounded + "\t");
                }
            logFile.println();
        }
        logFile.close();
    }

    /**
     * Defines the header of the MCMC output file. Some interface programs (e.g., R) have restrictions on the name
     * of the variables. If we set a header that can be interpreted as variable names by e.g. R, then it will be easier
     * to load these files in the respective environments.
     */

    public static void setEdgeFormat(int format) {
        edgeFormat = format;
    }

    protected static String formatEdge(String x, String y, int type) {
        if (edgeFormat == EDGE_FORMAT_R) {
            if (type == 1)
                return x + ".D." + y;
            else
                return x + ".B." + y;
        }
        if (type == 1)
            return x + " -> " + y;
        else
            return x + " <-> " + y;
    }
}

