package rbas.dmg.app;

import java.util.*;

import edu.cmu.tetrad.graph.*;
import edu.cmu.tetrad.data.*;
import edu.cmu.tetrad.util.MatrixUtils;

import java.io.*;

import rbas.dmg.sem.McmcProbitMixedGraph;

/**
 * RunGibbsGaussian
 *
 * Application for generating samples from the posterior of a Gaussian mixed graph model. Optionally, it
 * also computes predictive log-likelihood in a test set.
 *
 * Currently not supporting cyclic graphs.
 *
 * @author Ricardo Silva
 * @version $Revision: 0.1 $ $Date: 2007/07/05$
 */

public class RunGibbsProbit {

    protected static final boolean PRINT_CONTINGENCY_TABLE = true;

    public static void main(final String[] argv) {
        System.out.println();
        System.out.println("** DMG MODELING SOFTWARE: GIBBS SAMPLING FOR PROBIT MODELS");
        System.out.println("Code: Ricardo Silva, Statslab, University of Cambridge");
        System.out.println("Version: 0.2 - October 2007");

        if (argv.length != 4) {
            System.out.println("Usage: rbas.dmg.app.RunGibbsProbit <domain_name> <#samples> <#burn in> <step size>");
            System.out.println("Requires: files domain_name.graph, domain_name.train, domain_name.priors, domain_name.names");
            System.out.println("Optional: file domain_name.test");
        }

        try {
            System.out.println("Loading... " + argv[0] + ".train");
            RectangularDataSet data = Loader.loadDataSet(argv[0] + ".train");
            Graph graph = Loader.loadGraph(argv[0] + ".graph", data.getVariableNames());
            rbas.util.RbasGraphUtils.printEdges(graph);
            double vPrior[][] = new double[graph.getNodes().size()][graph.getNodes().size()];
            double df[] = new double[1];
            double bPrior[][][] = new double[graph.getNodes().size()][][];
            boolean fixedB[][] = new boolean[graph.getNodes().size()][];
            double values[][] = Loader.loadTypes(argv[0] + ".names", graph);
            double thresholdPriors[][][] = new double[graph.getNodes().size()][][];
            boolean thresholdFixed[][] = new boolean[graph.getNodes().size()][];
            for (int i = 0; i < graph.getNumNodes(); i++) {
                if (values[i] != null) {
                    thresholdPriors[i] = new double[values[i].length - 1][2];
                    thresholdFixed[i] = new boolean[values[i].length - 1];
                    for (int j = 0; j < thresholdFixed[i].length; j++)
                        thresholdFixed[i][j] = true; //TODO: REMOVE THIS
                }
            }
            Loader.loadPriors(argv[0] + ".priors", vPrior, df, bPrior, fixedB, thresholdPriors, thresholdFixed, graph);
            McmcProbitMixedGraph sampler = new McmcProbitMixedGraph(graph, data, vPrior, df[0], bPrior, fixedB,
                                             thresholdPriors, thresholdFixed, values);
            inference(sampler, Integer.parseInt(argv[1]), Integer.parseInt(argv[2]), Integer.parseInt(argv[3]), argv[0], null);
            //multiInference(sampler, Integer.parseInt(argv[1]), Integer.parseInt(argv[2]), Integer.parseInt(argv[3]), 80, argv[0]);
        } catch (IOException e) {
            System.out.println("File not loaded!");
            System.exit(0);
        }
	}

    public static void multiInference(McmcProbitMixedGraph mcmc, int simulationSize, int burnIn, int step,
                                      int numChains, String fileName) throws java.io.IOException {
        double results[] = new double[numChains];
        List   samplesHere = new ArrayList();
        for (int i = 0; i < numChains; i++) {
            System.out.println(">>>>>>>>>>>>> CHAIN #" + (i + 1));
            results[i] = inference(mcmc, simulationSize, burnIn, step, fileName, samplesHere);
            System.out.println("Saving MCMC run...");
            try {
               dumpSamplingObs(fileName, samplesHere, simulationSize, burnIn, step, mcmc, "." + (i + 1));
            } catch (java.io.IOException e) {
               throw new RuntimeException(e);
            }
        }
        System.out.println("All MCMC runs: ");
        for (int i = 0; i < numChains; i++)
            System.out.println(results[i]);
        System.out.println("Done!");
    }
    
    public static double inference(McmcProbitMixedGraph mcmc, int simulationSize, int burnIn, int step, String fileName,
                                   List samplesHere) throws java.io.IOException {
        Date date1 = new Date();
        List samples = mcmc.doSample(simulationSize);
        Date date2 = new Date();
        System.out.println("Started at " + date1.toString() + ",finished at " + date2.toString());

        if (samplesHere != null) {
            samplesHere.clear();
            samplesHere.addAll(samples);
        }

        System.out.println("Dumping samples...");
        dumpSamplingObs(fileName, samples, simulationSize, burnIn, step, mcmc, "");

        if (RunGibbsProbit.PRINT_CONTINGENCY_TABLE)
            mcmc.impliedContingencyTable(samples, burnIn, simulationSize, step);


//        RectangularDataSet test = Loader.loadDataSet(fileName + ".test");
//        System.out.println("-----------------------------------------------------");
//        System.out.println("Predictive loglikelihood: ");
//        double ll = mcmc.predictiveLoglikelihood(samples, test, burnIn, simulationSize, step);
//        System.out.println(ll);
//        System.out.print("Predictive loglikelihood per data point: ");
//        System.out.println(ll / test.getNumRows());

        System.out.println("Done!");

//        return ll;
        return -1;
    }

    protected static void dumpSampling(String fileName, List samples,int simulationSize, int burnIn,
                                       int step, McmcProbitMixedGraph mcmc) throws IOException {
        Graph graph = mcmc.getGraph();
        PrintWriter logFile = new PrintWriter(fileName + ".mcmc");
        Graph bidirectedGraph = rbas.util.RbasGraphUtils.getBidirectedSubgraph(graph);
        int numNodes = graph.getNumNodes();

        List<String> directedEdges = new ArrayList<String>();   //edges need to reordered according to order of parameters in 'samples'
        List<String> bidirectedEdges = new ArrayList<String>(); //edges need to reordered according to order of parameters in 'samples'
        int preSpouses[][] = new int[numNodes][];

        int bidirAdj[][] = new int[numNodes][numNodes];
        for (int i = 0; i < numNodes; i++) {
            Node node1 = graph.getNodes().get(i);
            bidirAdj[i][i] = 1;
            int numPreSpouses = 0;
            for (int j = 0; j < numNodes; j++) {
                Node node2 = graph.getNodes().get(j);
                if (graph.isParentOf(node2, node1))
                    directedEdges.add(node2 + " -> " + node1);
                if ((j < i && bidirectedGraph.isAdjacentTo(node1, node2)) || (j == i)) {
                    bidirectedEdges.add(node1.getName() + " <-> " + node2.getName());
                    bidirAdj[i][j] = 1;
                    numPreSpouses++;
                }
            }
            directedEdges.add("_INT -> " + node1);
            preSpouses[i] = new int[numPreSpouses];
            int psCount = 0;
            for (int j = 0; j < i; j++)
                if (bidirAdj[i][j] > 0)
                    preSpouses[i][psCount++] = j;
            preSpouses[i][numPreSpouses - 1] = i;
        }
        for (String edgeStr : directedEdges)
            logFile.print(edgeStr + "\t");
        for (String edgeStr : bidirectedEdges)
            logFile.print(edgeStr + "\t");
        logFile.println();
        double allBidirected[][][] = (double[][][]) samples.get(0);
        double allDirected[][][] = (double[][][]) samples.get(2);
        for (int s = burnIn; s < simulationSize; s += step) {
            double errSample[][] = mcmc.unpackCovariance(allBidirected[s]);
            double coeffSample[][] = allDirected[s];
            for (int i = 0; i < numNodes; i++)
                for (int p = 0; p < coeffSample[i].length; p++)
                    logFile.print(coeffSample[i][p] + "\t");
            for (int i = 0; i < numNodes; i++)
                for (int p = 0; p < preSpouses[i].length; p++)
                    logFile.print(errSample[i][preSpouses[i][p]] + "\t");
            logFile.println();
        }
        logFile.close();
    }

    protected static void dumpSamplingObs(String fileName, List samples, int simulationSize, int burnIn,
                                          int step, McmcProbitMixedGraph mcmc, String label) throws IOException {
        PrintWriter logFile = new PrintWriter(fileName + ".obs" + label + ".mcmc");

        int      numObserved = mcmc.getNumObserved();
        double ordValues[][] = mcmc.getOrdValues();
        int  observedIndex[] = mcmc.getObservedIndex();
        int         numCat[] = mcmc.getNumCat();

        int entryHeader[] = new int[numObserved];
        for (int i = 0; i < numObserved; i++)
             entryHeader[i] = 0;
        nextEntryHeader:
        while (true) {
            logFile.print("P");
            for (int i = 0; i < numObserved; i++) {
                logFile.print(entryHeader[i]);
                if (i < numObserved - 1)
                    logFile.print(".");
            }
            logFile.print("\t");
            //Move to the next combination of values
            boolean carry = false;
            for (int i = numObserved - 1; i >= 0; i--) {
                if (i == numObserved - 1)
                    entryHeader[i]++;
                if (carry)
                    entryHeader[i]++;
                carry = (entryHeader[i] >= numCat[observedIndex[i]]);
                entryHeader[i] = entryHeader[i] % numCat[observedIndex[i]];
            }

            for (int i = 0; i < numObserved; i++)
                if (entryHeader[i] != 0)
                    continue nextEntryHeader;
            break;
        }
        logFile.println();

        double allBidirected[][][] = (double[][][]) samples.get(0);
        double allDirected[][][] = (double[][][]) samples.get(2);
        double allThresholds[][][] = (double[][][]) samples.get(3);

        double singleBidirected[][][] = new double[1][][];
        double singleDirected[][][] = new double[1][][];
        double singleThresholds[][][] = new double[1][][];

        List singleSample = new ArrayList();
        singleSample.add(singleBidirected); singleSample.add(null);
        singleSample.add(singleDirected); singleSample.add(singleThresholds);

        for (int s = burnIn; s < simulationSize; s += step) {
            List<double[]> entries = new ArrayList<double[]>();
            double entry[] = new double[numObserved + 1];
            for (int i = 0; i < numObserved; i++)
                 entry[i] = 0.;

            nextEntry:
            while (true) {

                double pattern[] = new double[numObserved];
                for (int i = 0; i < numObserved; i++)
                    pattern[i] = ordValues[observedIndex[i]][(int) entry[i]];

                singleBidirected[0] = allBidirected[s];
                singleDirected[0] = allDirected[s];
                singleThresholds[0] = allThresholds[s];

                entry[numObserved] = mcmc.evaluateProbability(singleSample, burnIn, pattern);
                entries.add(entry);

                //Move to the next combination of values
                double oldEntry[] = entry;
                entry = new double[numObserved + 1];
                System.arraycopy(oldEntry, 0, entry, 0, numObserved);
                boolean carry = false;
                for (int i = numObserved - 1; i >= 0; i--) {
                    if (i == numObserved - 1)
                        entry[i]++;
                    if (carry)
                        entry[i]++;
                    carry = (entry[i] >= numCat[observedIndex[i]]);
                    entry[i] = entry[i] % numCat[observedIndex[i]];
                }

                for (int i = 0; i < numObserved; i++)
                    if (entry[i] != 0)
                        continue nextEntry;
                break;
            }

            for (double e[] : entries) {
                double rounded = ((int) (e[numObserved] * 100000)) / 100000.;
                //double rounded = e[numObserved];
                logFile.print(rounded + "\t");
            }
            logFile.println();

        }

        logFile.close();

    }

    protected static void dumpSamplingObs2(String fileName, List samples, int simulationSize, int burnIn,
                                           int step, McmcProbitMixedGraph mcmc, String label) throws IOException {
        Graph graph = mcmc.getGraph();
        PrintWriter logFile = new PrintWriter(fileName + ".obs" + label + ".mcmc");

        List<String> name = new ArrayList<String>();
        for (Node node : graph.getNodes())
            if (node.getNodeType() == NodeType.MEASURED)
                name.add(node.getName());
        Object nameArray[] = name.toArray();
        java.util.Arrays.sort(nameArray);
        Hashtable<String, Integer> namePos = new Hashtable<String, Integer>();
        for (int g = 0; g < graph.getNumNodes(); g++) {
            Node node = graph.getNodes().get(g);
            if (node.getNodeType() == NodeType.LATENT)
                continue;
            for (int i = 0; i < nameArray.length; i++)
                if (nameArray[i].equals(node.getName())) {
                    namePos.put((String) nameArray[i], new Integer(g));
                    break;
                }
        }


        Graph bidirectedGraph = rbas.util.RbasGraphUtils.getBidirectedSubgraph(graph);
        int numNodes = graph.getNumNodes(), numObsNodes = name.size();

        int preSpouses[][] = new int[numNodes][];
        int bidirAdj[][] = new int[numNodes][numNodes];
        for (int i = 0; i < numNodes; i++) {
            Node node1 = graph.getNodes().get(i);
            bidirAdj[i][i] = 1;
            int numPreSpouses = 0;
            for (int j = 0; j < numNodes; j++) {
                Node node2 = graph.getNodes().get(j);
                if ((j < i && bidirectedGraph.isAdjacentTo(node1, node2)) || (j == i)) {
                    bidirAdj[i][j] = 1;
                    numPreSpouses++;
                }
            }
            preSpouses[i] = new int[numPreSpouses];
            int psCount = 0;
            for (int j = 0; j < i; j++)
                if (bidirAdj[i][j] > 0)
                    preSpouses[i][psCount++] = j;
            preSpouses[i][numPreSpouses - 1] = i;
        }
        for (int i = 0; i < numObsNodes; i++)
            for (int j = i; j < numObsNodes; j++)
                logFile.print("X" + (i + 1) + ".X" + (j + 1) + "\t");
        logFile.println();
        double allBidirected[][][] = (double[][][]) samples.get(0);
        double allDirected[][][] = (double[][][]) samples.get(2);
        double iMinusB[][] = new double[numNodes][numNodes];
        int    parents[][] = rbas.util.RbasGraphUtils.getParentsMatrix(graph);
        for (int s = burnIn; s < simulationSize; s += step) {
            double errSample[][] = mcmc.unpackCovariance(allBidirected[s]);
            double coeffSample[][] = allDirected[s];
            for (int i = 0; i < numNodes; i++)
                for (int j = 0; j < numNodes; j++)
                    iMinusB[i][j] = 0.;
            for (int i = 0; i < numNodes; i++)
                iMinusB[i][i] = 1.;
            for (int i = 0; i < numNodes; i++)
                for (int j = 0; j < parents[i].length; j++)
                    iMinusB[i][parents[i][j]] -= coeffSample[i][j];
            double invIMinusB[][] = MatrixUtils.inverse(iMinusB);
            double implCov[][] = MatrixUtils.product(invIMinusB,
                                     MatrixUtils.product(errSample, MatrixUtils.transpose(invIMinusB)));
            for (int i = 0; i < numObsNodes; i++)
                for (int j = i; j < numObsNodes; j++) {
                    double rounded = ((int) (implCov[namePos.get((String)nameArray[i])][namePos.get((String)nameArray[j])]
                                     * 1000)) / 1000.;
                    logFile.print(rounded + "\t");
                }
            logFile.println();
        }
        logFile.close();
    }

}

