package rbas.dmg.sem;

import edu.cmu.tetrad.graph.*;
import edu.cmu.tetrad.util.MatrixUtils;
import edu.cmu.tetrad.util.ProbUtils;
import edu.cmu.tetrad.data.ColtDataSet;
import edu.cmu.tetrad.data.RectangularDataSet;
import edu.cmu.tetrad.data.ContinuousVariable;
import rbas.util.RbasGraphUtils;
import rbas.util.RbasProbUtils;

import java.util.*;
import java.util.logging.Logger;
import java.util.logging.StreamHandler;
import java.util.logging.Level;
import java.util.logging.Handler;

/**
 * McmcGaussianMixedGraph
 *
 * Implements a Markov Chain Monte Carlo sampler for a Gaussian mixed graph model, i.e.,
 * a graph with directed and bi-directed edges (possibly two edges per pair of nodes).
 *
 * There are two choices of samplers: one of them is a purely Gibbs sampling procedure, useful
 * to obtain posterior distributions of parameters. The other is a combination of Gibbs +
 * importance sampler, useful for computing expectations.
 *
 * Currently not supporting cyclic graphs.
 *
 * @author Ricardo Silva
 * @version $Revision: 1.1 $ $Date: 2007/09/21$
 */


public class McmcGaussianMixedGraph {

    public static Logger logger = Logger.getLogger("rbas.dmg.sem.McmcGaussianMixedGraph");

    double  bPrior[][][], vPrior[][], vPriorDf;
    double  yCentered[][], iMinusB[][];
    double  rawData[][];
    int     sampleSize, numNodes, numObserved, numLatents;
    int     numPreSpouses[][], numNotPreSpouses[][], numSpouses[][], numNotSpouses[][], numPosSpouses[][];
    int     preSpouses[][][], notPreSpouses[][][], spouses[][][], notSpouses[][][], posSpouses[][][];
    int     parents[][], children[][];
    boolean bidirectedMatrix[][], isLatent[], fixedParams[][];
    Graph   graph, bidirectedGraph;
    boolean isAcyclic;
    RectangularDataSet train;

    double eLatentOthers[][], eOthersOthers[][][], eLatentLatent, eNodes[], eeNodes[][], regress[][], centerMoments[][];
    double residualVar[];
    double eObservedLatents[][], eObservedObserved[][], latentsResidualCov[][], latentRegress[][];
    double latentBuffer[], latentMean[];

    int    latentIndex[], latentIndexReverse[], observedIndex[], observedIndexReverse[], blanket[][];

    int       numDistricts, districts[][];
    int       districtNeighbors[][];
    double    invBuffer[][][];
    Hashtable<Node, Integer> nodeIndex;

    protected boolean isDag;
    protected boolean blockLatentSampling, returnLatents = false;

    double eeCoeff[][][], eCoeff[][], covCoeff[][][], cBuffer1[][], cBuffer2[][];
    int    indexCoeff[][][]; //for each district [d], each free coefficient [d][c] is mapped to its child [d][c][0] and
                             //and its position with respect to other parents of h (index: [d][c][1])
    double eeData[][], eData[];
    double priorCovCoeff[][], priorMeanCovCoeff[][];

    RectangularDataSet reorderedData;

    public McmcGaussianMixedGraph(Graph graph, RectangularDataSet train, double vPrior[][], double df,
                                  double bPrior[][][], boolean fixedB[][]) {
        this.isAcyclic = !graph.existsDirectedCycle();
        if (!this.isAcyclic)
            throw new RuntimeException("Cyclic graphics not currently supported!");
        this.train = train;
        List factorization = RbasGraphUtils.factorizeGraph(graph, false);
        this.graph = (Graph) factorization.get(0);
        this.districts = (int[][])  factorization.get(1);
        this.districtNeighbors = (int[][])  factorization.get(2);
        this.numNodes = this.graph.getNumNodes();
        this.numDistricts = this.districts.length;
        startPriors(vPrior, bPrior, fixedB, graph);
        this.vPriorDf = df;
        initialize();
    }

    /****************************************************************************************************************
     * INITIALIZATION: setup all data structures. Re-order the data to correspond to the node order in the graph.
     * Make sure to include "dummy" columns for possible latent variables.
     */

    /**
     * Set up all prior arrays that have to be created, including those related to fixed (i.e. no random parameters).
     * Notice that because of the reordering of nodes as done in rebuildOrderedGraph, we can't just make a direct
     * copy of the priors given as arguments to the constructors. This is the reason why we build a mapping of nodes
     * in the original order to the new order.
     */

    protected void startPriors(double vPrior[][], double bPrior[][][], boolean fixedB[][], Graph oldGraph) {
        List nodes = this.graph.getNodes();
        List originalNodes = oldGraph.getNodes();
        int map[] = new int[nodes.size()];
        for (int i = 0; i < originalNodes.size(); i++) {
            int pos = -1;
            for (int j = 0; j < nodes.size(); j++)
                if (nodes.get(j) == originalNodes.get(i)) {
                    pos = j;
                    break;
                }
            map[i] = pos;
        }
        this.vPrior = new double[nodes.size()][nodes.size()];
        this.bPrior = new double[nodes.size()][][];
        this.fixedParams = new boolean[nodes.size()][];
        for (int i = 0; i < originalNodes.size(); i++) {
            for (int j = 0; j < originalNodes.size(); j++)
                 this.vPrior[map[i]][map[j]] = vPrior[i][j];
            this.bPrior[map[i]] = new double[bPrior[i].length][];
            this.fixedParams[map[i]] = new boolean[fixedB[i].length];
            List oldParents = oldGraph.getParents((Node) originalNodes.get(i));
            List parents = this.graph.getParents((Node) originalNodes.get(i));
            for (int j = 0; j < oldParents.size(); j++) {
                int pos = -1;
                for (int k = 0; k < parents.size(); k++)
                    if (oldParents.get(j) == parents.get(k)) {
                        pos = k;
                        break;
                    }
                this.bPrior[map[i]][pos] = bPrior[i][j];
                this.fixedParams[map[i]][pos] = fixedB[i][j];
            }
            //*IMPORTANT*: Notice that the last parent is always the intercept term (which will be set to 1 in the data)
            this.bPrior[map[i]][oldParents.size()] = bPrior[i][oldParents.size()];
            this.fixedParams[map[i]][oldParents.size()] = fixedB[i][oldParents.size()];
        }
    }

    /**
     * Allocate memory for all data structures that are necessary, and generate indexes to map each node to its
     * respective district, as well as its Markov blanket.
     */

    protected void initialize() {
        //Part I: setup the data. Columns should correspond to the same order as in the original graph,
        //plus have dummy columns for the latent variables

        this.sampleSize = this.train.getNumRows();
        List<Node> columnNodes = new ArrayList<Node>();
        for (int v = 0; v < this.numNodes; v++)
            columnNodes.add(new ContinuousVariable(this.graph.getNodes().get(v).toString()));
        columnNodes.add(new ContinuousVariable("___intercept"));
        this.reorderedData =  new ColtDataSet(this.sampleSize, columnNodes);
        this.numLatents = this.numObserved = 0;
        this.isLatent = new boolean[this.graph.getNumNodes() + 1];
        this.isLatent[this.graph.getNumNodes()] = false;

        List<String> variableNames = this.train.getVariableNames();
        for (int i = 0; i < this.numNodes; i++) {
            if (this.graph.getNodes().get(i).getNodeType() == NodeType.LATENT) {
                this.isLatent[i] = true;
                this.numLatents++;
            }
            else {
                int col = -1;
                for (int t = 0; t < this.numNodes; t++)
                     if (variableNames.get(t).equals(this.graph.getNodes().get(i).getName())) {
                         col = t;
                         break;
                     }
                this.isLatent[i] = false;
                for (int s = 0; s < this.sampleSize; s++)
                     this.reorderedData.setDouble(s, i, this.train.getDouble(s,col));
                this.numObserved++;
            }
        }
        int latentCount = 0, observedCount = 0;
        this.latentIndex = new int[this.numLatents]; this.latentIndexReverse = new int[this.numNodes];
        this.observedIndex = new int[this.numObserved]; this.observedIndexReverse = new int[this.numNodes];
        for (int i = 0; i < this.numNodes; i++) {
            if (this.graph.getNodes().get(i).getNodeType() == NodeType.LATENT) {
                this.latentIndexReverse[i] = latentCount;
                this.latentIndex[latentCount++] = i;
            }
            else {
                this.observedIndexReverse[i] = observedCount;
                this.observedIndex[observedCount++] = i;
            }
        }
        this.rawData = this.reorderedData.getDoubleData().toArray();
        for (int i = 0; i < this.sampleSize; i++)
            this.rawData[i][this.numNodes] = 1.;  //Pseudo-data for the intercept term

        //Part II: initialize graph data structures

        this.bidirectedGraph = RbasGraphUtils.getBidirectedSubgraph(this.graph);
        this.bidirectedMatrix = new boolean[this.numNodes][this.numNodes];
        this.preSpouses = new int[this.districts.length][][];     //*ordered* preSpouses, non-preSpouses, etc.
        this.notPreSpouses = new int[this.districts.length][][];  //indexed by districts
        this.posSpouses = new int[this.districts.length][][];
        this.spouses = new int[this.districts.length][][];
        this.notSpouses = new int[this.districts.length][][];
        this.numPreSpouses = new int[this.districts.length][];
        this.numNotPreSpouses = new int[this.districts.length][];
        this.numPosSpouses = new int[this.districts.length][];
        this.numSpouses = new int[this.districts.length][];
        this.numNotSpouses = new int[this.districts.length][];
        for (int i = 0; i < this.numNodes; i++) {
            for (int j = 0; j < i; j++) {
                if (this.bidirectedGraph.isAdjacentTo(this.bidirectedGraph.getNodes().get(i),
                    this.bidirectedGraph.getNodes().get(j))) {
                    this.bidirectedMatrix[i][j] = this.bidirectedMatrix[j][i] = true;
                }
                else {
                    this.bidirectedMatrix[i][j] = false;
                }
            }
        }
        this.invBuffer = new double[this.numDistricts][][];
        for (int d = 0; d < this.districts.length; d++) {
            this.invBuffer[d] = new double[this.districts[d].length][this.districts[d].length];
            this.preSpouses[d] = new int[this.districts[d].length][];
            this.notPreSpouses[d] = new int[this.districts[d].length][];
            this.posSpouses[d] = new int[this.districts[d].length][];
            this.spouses[d] = new int[this.districts[d].length][];
            this.notSpouses[d] = new int[this.districts[d].length][];
            this.numPreSpouses[d] = new int[this.districts[d].length];
            this.numNotPreSpouses[d] = new int[this.districts[d].length];
            this.numPosSpouses[d] = new int[this.districts[d].length];
            this.numSpouses[d] = new int[this.districts[d].length];
            this.numNotSpouses[d] = new int[this.districts[d].length];
            for (int i = 0; i < this.districts[d].length; i++) {
                this.numPreSpouses[d][i] = this.numNotPreSpouses[d][i] = this.numSpouses[d][i] = this.numPosSpouses[d][i] = 0;
                for (int j = 0; j < i; j++) {
                    if (this.bidirectedMatrix[this.districts[d][i]][this.districts[d][j]]) {
                        this.numPreSpouses[d][i]++;
                        this.numPosSpouses[d][j]++;
                        this.numSpouses[d][i]++; this.numSpouses[d][j]++;
                    }
                    else {
                        this.numNotPreSpouses[d][i]++;
                        this.numNotSpouses[d][i]++; this.numNotSpouses[d][j]++;
                    }
                }
                this.preSpouses[d][i] = new int[this.numPreSpouses[d][i]];
                this.notPreSpouses[d][i] = new int[this.numNotPreSpouses[d][i]];
                int countS = 0, countNS = 0;
                for (int j = 0; j < i; j++) {
                    if (this.bidirectedMatrix[this.districts[d][i]][this.districts[d][j]])
                        this.preSpouses[d][i][countS++] = j;
                    else
                        this.notPreSpouses[d][i][countNS++] = j;
                }
            }
            for (int i = 0; i < this.districts[d].length; i++) {
                this.spouses[d][i] = new int[this.numSpouses[d][i]];
                this.notSpouses[d][i] = new int[this.numNotSpouses[d][i]];
                this.posSpouses[d][i] = new int[this.numPosSpouses[d][i]];
                int count = 0, countNot = 0, countPos = 0;
                for (int j = 0; j < this.districts[d].length; j++)
                    if (this.bidirectedMatrix[this.districts[d][i]][this.districts[d][j]] && i != j)
                        this.spouses[d][i][count++] = j;
                    else if (i != j)
                        this.notSpouses[d][i][countNot++] = j;
                for (int j = i + 1; j < this.districts[d].length; j++)
                    if (this.bidirectedMatrix[this.districts[d][j]][this.districts[d][i]])
                        this.posSpouses[d][i][countPos++] = j;
            }
        }
        this.parents = new int[this.numNodes][];
        for (int i = 0; i < this.numNodes; i++) {
            List parents = this.graph.getParents((Node) this.graph.getNodes().get(i));
            this.parents[i] = new int[parents.size() + 1];
            this.parents[i][parents.size()] = this.numNodes; //The intercept term
            for (int j = 0; j < parents.size(); j++) {
                Node parent = (Node) parents.get(j);
                for (int k = 0; k < this.graph.getNodes().size(); k++) {
                    Node node = (Node) this.graph.getNodes().get(k);
                    if (node == parent) {
                        this.parents[i][j] = k;
                        break;
                    }
                }
            }
        }

        this.nodeIndex = new Hashtable<Node, Integer>();
        for (int i = 0; i < this.numNodes; i++)
            this.nodeIndex.put(this.graph.getNodes().get(i), i);
        this.blanket = RbasGraphUtils.getMarkovBlanketsDmg(this.graph);
        this.children = RbasGraphUtils.getChildrenMatrix(this.graph);

        this.isDag = true;
        for (Edge edge : this.graph.getEdges())
            if (edge.getEndpoint1() == Endpoint.ARROW && edge.getEndpoint2() == Endpoint.ARROW) {
                this.isDag = false;
                break;
            }

        //Part III: initialize regression data structures (for latent variable sampling)

        this.yCentered = new double[this.numNodes][this.numNodes];
        this.iMinusB = new double[this.numNodes][this.numNodes];

        this.eLatentOthers = new double[this.numLatents][];
        this.eOthersOthers = new double[this.numLatents][][];
        for (int i = 0; i < this.numLatents; i++) {
            int latentIdx = this.latentIndex[i];
            this.eLatentOthers[i] = new double[this.blanket[latentIdx].length + 1];
            this.eOthersOthers[i] = new double[this.blanket[latentIdx].length + 1][this.blanket[latentIdx].length + 1];
        }
        this.residualVar = new double[this.numLatents];
        this.regress = new double[this.numLatents][this.numObserved + 1];
        this.eNodes = new double[this.numNodes];
        this.eeNodes = new double[this.numNodes][this.numNodes];
        this.centerMoments = new double[this.numNodes][this.numNodes];

        this.blockLatentSampling = true;
        this.latentBuffer = new double[this.numLatents];
        this.latentMean = new double[this.numLatents];
        this.eObservedLatents = new double[this.numObserved + 1][this.numLatents];
        this.eObservedObserved =  new double[this.numObserved + 1][this.numObserved + 1];
        this.latentsResidualCov = new double[this.numLatents][this.numLatents];
        this.latentRegress = new double[this.numLatents][this.numObserved + 1];

        cacheMoments(this.rawData);
        createCoefficientIndex();
    }

    protected void cacheMoments(double data[][]) {
        this.eeData = new double[this.numNodes + 1][this.numNodes + 1];
        this.eData = new double[this.numNodes + 1];
        for (int d = 0; d < data.length; d++)
            for (int i = 0; i < this.numNodes + 1; i++) {
                for (int j = i; j < this.numNodes + 1; j++)
                    this.eeData[i][j] += data[d][i] * data[d][j];
                this.eData[i] += data[d][i];
            }
        for (int i = 0; i < this.numNodes + 1; i++)
            for (int j = i + 1; j < this.numNodes + 1; j++)
                this.eeData[j][i] = this.eeData[i][j];
    }

    protected void updateCacheMoments(double data[][]) {
        if (this.numLatents == 0)
            return;
        for (int i = 0; i < this.numNodes + 1; i++) {
            for (int j = i; j < this.numNodes + 1; j++)
                if (this.isLatent[i] || this.isLatent[j])
                    this.eeData[i][j] = 0.;
            if (this.isLatent[i])
                this.eData[i] = 0.;
        }
        for (int d = 0; d < data.length; d++)
            for (int i = 0; i < this.numNodes + 1; i++) {
                for (int j = i; j < this.numNodes + 1; j++)
                    if (this.isLatent[i] || this.isLatent[j])
                        this.eeData[i][j] += data[d][i] * data[d][j];
                if (this.isLatent[i])
                    this.eData[i] += data[d][i];
            }
        for (int i = 0; i < this.numNodes + 1; i++)
            for (int j = i + 1; j < this.numNodes + 1; j++)
                if (this.isLatent[i] || this.isLatent[j])
                    this.eeData[j][i] = this.eeData[i][j];
    }

    protected void createCoefficientIndex() {
        this.eeCoeff = new double[this.numDistricts][][];
        this.eCoeff = new double[this.numDistricts][];
        this.covCoeff = new double[this.numDistricts][][];
        this.cBuffer1 = new double[this.numDistricts][]; this.cBuffer2 = new double[this.numDistricts][];
        this.indexCoeff = new int[this.numDistricts][][];

        this.priorCovCoeff = new double[this.numDistricts][];
        this.priorMeanCovCoeff = new double[this.numDistricts][];

        for (int d = 0; d < this.numDistricts; d++) {
            int numFreeParams = 0;
            for (int v = 0; v < this.districts[d].length; v++) {
                int vv = this.districts[d][v];
                for (int p = 0; p < this.parents[vv].length; p++)
                    if (!this.fixedParams[vv][p])
                        numFreeParams++;
            }
            this.eeCoeff[d] = new double[numFreeParams][numFreeParams];
            this.eCoeff[d] = new double[numFreeParams];
            this.covCoeff[d] = new double[numFreeParams][numFreeParams];
            this.cBuffer1[d] = new double[numFreeParams]; this.cBuffer2[d] = new double[numFreeParams];
            this.indexCoeff[d] = new int[numFreeParams][2];
            this.priorCovCoeff[d] = new double[numFreeParams];
            this.priorMeanCovCoeff[d] = new double[numFreeParams];

            int pos = 0;
            for (int v = 0; v < this.districts[d].length; v++) {
                int vv = this.districts[d][v];
                for (int p = 0; p < this.parents[vv].length; p++)
                    if (!this.fixedParams[vv][p]) {
                        this.indexCoeff[d][pos][0] = vv;
                        this.indexCoeff[d][pos][1] = p;
                        this.priorCovCoeff[d][pos] = 1. / this.bPrior[vv][p][1];
                        this.priorMeanCovCoeff[d][pos] = this.bPrior[vv][p][0] / this.bPrior[vv][p][1];
                        pos++;
                    }
            }
        }
    }

    /**
     * Sampling methods
     */

    public List doSample(int number) {
        logger.fine("*** START SAMPLING");
        logger.fine("");

        Random r = new Random();
        double samplesBidirected[][][]    = new double[number][][];
        double samplesDirected[][][]      = new double[number][][];
        double samplesWeight[]            = new double[number];
        double directed[][]               = new double[this.numNodes][];
        double bidirected[][]             = null;
        double latents[][][]              = null;

        if (this.returnLatents)
            latents = new double[number][this.sampleSize][this.numLatents];

        boolean gibbsMode = true;

        //Initialization
        for (int i = 0; i < this.numNodes; i++) {
            directed[i] = new double[this.parents[i].length];
            for (int p = 0; p < this.parents[i].length; p++) {
                if (!this.fixedParams[i][p])
                    directed[i][p] = r.nextGaussian();
                else
                    directed[i][p] = this.bPrior[i][p][0];
            }
        }
        if (gibbsMode) {
            bidirected = new double[this.numNodes][this.numNodes];
            for (int i = 0; i < this.numNodes; i++)
                bidirected[i][i] = 1. + r.nextDouble();
        }
        for (int i = 0; i < this.numNodes; i++)
            if (this.isLatent[i])
                for (int j = 0; j < this.sampleSize; j++)
                    this.rawData[j][i] = r.nextGaussian();
        updateCacheMoments(this.rawData);
        double impliedGInvWishartParam[][] = getImpliedGInvWishartParam(directed);

        //Iteration
        for (int i = 0; i < number; i++) {
            logger.finer("- Sample #" + (i + 1));

            List<double[][]> matrices;
            if (gibbsMode)
                matrices = sampleBidirectedMatrixGibbs(r, this.vPriorDf + this.sampleSize, impliedGInvWishartParam, bidirected);
            else
                matrices = sampleBidirectedMatrixImportance(r, this.vPriorDf + this.sampleSize, impliedGInvWishartParam);
            samplesBidirected[i] = matrices.get(0);
            double samplesBidirectedInv[][] = matrices.get(1), weight = matrices.get(2)[0][0];
            samplesDirected[i] = sampleDirectedMatrixBlock(r, samplesBidirected[i], samplesBidirectedInv, directed);
            directed = samplesDirected[i]; bidirected = samplesBidirected[i];
            if (this.numLatents > 0) {
                if (this.blockLatentSampling)
                    sampleLatentsBlock(r, samplesBidirected[i], samplesDirected[i], this.rawData);
                else
                    sampleLatents(r, samplesBidirected[i], samplesDirected[i], this.rawData);
                updateCacheMoments(this.rawData);
                if (this.returnLatents)
                    for (int s = 0; s < this.sampleSize; s++)
                        for (int l = 0; l < this.numLatents; l++)
                            latents[i][s][l] = this.rawData[s][this.latentIndex[l]];
            }
            impliedGInvWishartParam = getImpliedGInvWishartParam(samplesDirected[i]);
            samplesWeight[i] = weight;
            samplesBidirected[i] = packCovariance(samplesBidirected[i]);
        }

        List samples = new ArrayList();
        samples.add(samplesBidirected); samples.add(samplesWeight); samples.add(samplesDirected);
        if (this.returnLatents)
            samples.add(latents);
        return samples;
    }

    /**
     * Sample latents given parameters and data. A given latents are sampled by conditioning on its Markov blanket.
     * This might not converge as fast as sampling all latents simultaneously, but can be more efficient if
     * latents are sparsely connected.
     */

    protected void sampleLatents(Random r, double v[][], double b[][], double data[][]) {
        getImpliedMoments(v, b);
        for (int i = 0; i < this.numLatents; i++) {
            int latentIdx = this.latentIndex[i];
            getLatentGivenObserved(r, i, v, b);
            for (int n = 0; n < data.length; n++){
                data[n][latentIdx] = this.regress[i][this.blanket[latentIdx].length];
                for (int j = 0; j < this.blanket[latentIdx].length; j++)
                    data[n][latentIdx] += this.regress[i][j] * data[n][this.blanket[latentIdx][j]];
                data[n][latentIdx] += r.nextGaussian() * Math.sqrt(this.residualVar[i]);
            }
        }
    }

    /**
     * Sample latents given parameters and data. Latents are sampled simultaneously. This is probably the most
     * desirable scheme for models found in practice, where latents are all conditionally dependent given the
     * data - so no factorization is of much use. Sampling latents simultaneously is likely to guarantee a much
     * better mixing of the chain.
     */

    protected void sampleLatentsBlock(Random r, double v[][], double b[][], double data[][]) {
        getImpliedMoments(v, b);
        getLatentsGivenObserved(r, v, b);

        double cholesky[][] = MatrixUtils.cholesky(this.latentsResidualCov);

        for (int n = 0; n < data.length; n++) {
            for (int i = 0; i < this.numLatents; i++) {
                this.latentBuffer[i] = r.nextGaussian();
                this.latentMean[i] = this.latentRegress[this.numObserved][i];
                for (int j = 0; j < this.numObserved; j++)
                    this.latentMean[i] += this.latentRegress[j][i] * data[n][this.observedIndex[j]];
            }
            for (int i = 0; i < this.numLatents; i++) {
                int latentIdx = this.latentIndex[i];
                data[n][latentIdx] = this.latentMean[i];
                for (int j = 0; j < this.numLatents; j++)
                    data[n][latentIdx] += cholesky[i][j] * this.latentBuffer[j];
            }
        }
    }

    /**
     * Calculates implied first and second moments of all hidden variables given observed variables and parameters.
     */

    protected void getLatentGivenObserved(Random r, int latentNumber, double v[][], double b[][]) {
        int latentIdx = this.latentIndex[latentNumber];
        for (int j = 0; j < this.blanket[latentIdx].length; j++)
            this.eLatentOthers[latentNumber][j] = this.eeNodes[latentIdx][this.blanket[latentIdx][j]];
        this.eLatentOthers[latentNumber][this.blanket[latentIdx].length] = this.eNodes[latentIdx];
        this.eLatentLatent = this.eeNodes[latentIdx][latentIdx];

        for (int j = 0; j < this.blanket[latentIdx].length; j++) {
            for (int k = 0; k < this.blanket[latentIdx].length; k++)
                this.eOthersOthers[latentNumber][j][k] = this.eeNodes[this.blanket[latentIdx][j]][this.blanket[latentIdx][k]];
            this.eOthersOthers[latentNumber][j][this.blanket[latentIdx].length] =
                    this.eOthersOthers[latentNumber][this.blanket[latentIdx].length][j] =
                        this.eNodes[this.blanket[latentIdx][j]];
        }
        this.eOthersOthers[latentNumber][this.blanket[latentIdx].length][this.blanket[latentIdx].length] = 1;

        double invEYY[][] = MatrixUtils.inverse(this.eOthersOthers[latentNumber]);
        this.regress[latentNumber] = MatrixUtils.product(invEYY, this.eLatentOthers[latentNumber]);

        this.residualVar[latentNumber] = this.eLatentLatent - MatrixUtils.innerProduct(this.eLatentOthers[latentNumber],
                                                                                       this.regress[latentNumber]);
    }

    protected void getLatentsGivenObserved(Random r, double v[][], double b[][]) {
        for (int i = 0; i < this.numObserved; i++) {
            for (int j = 0; j < this.numLatents; j++)
                this.eObservedLatents[i][j] = this.eeNodes[this.observedIndex[i]][this.latentIndex[j]];
            for (int j = 0; j < this.numObserved; j++)
                this.eObservedObserved[i][j]= this.eeNodes[this.observedIndex[i]][this.observedIndex[j]];
            this.eObservedObserved[i][this.numObserved] = this.eObservedObserved[this.numObserved][i] =
                    this.eNodes[this.observedIndex[i]];
        }
        for (int i = 0; i < this.numLatents; i++)
            this.eObservedLatents[this.numObserved][i] = this.eNodes[this.latentIndex[i]];
        this.eObservedObserved[this.numObserved][this.numObserved] = 1.;

        double invEYY[][] = MatrixUtils.inverse(this.eObservedObserved);
        this.latentRegress = MatrixUtils.product(invEYY, this.eObservedLatents);
        double innerSum[][] = MatrixUtils.product(MatrixUtils.transpose(this.latentRegress),
                                                  MatrixUtils.product(this.eObservedObserved,
                                                          this.latentRegress));

        for (int i = 0; i < this.numLatents; i++)
            for (int j = i; j < this.numLatents; j++)
                this.latentsResidualCov[i][j] = this.latentsResidualCov[j][i] =
                        this.eeNodes[this.latentIndex[i]][this.latentIndex[j]] - innerSum[i][j];
    }

    protected void getImpliedMoments(double v[][], double directedParams[][]) {
        for (int i = 0; i < this.numNodes; i++) {
            for (int j = 0; j < this.numNodes; j++)
                this.iMinusB[i][j] = 0.;
            this.iMinusB[i][i] = 1.;
        }
        for (int i = 0; i < this.numNodes; i++)
            for (int p = 0; p < this.parents[i].length - 1; p++)
                this.iMinusB[i][this.parents[i][p]] -= directedParams[i][p];
        double inv[][] = MatrixUtils.inverse(this.iMinusB);
        for (int i = 0; i < this.numNodes; i++) {
            for (int j = i; j < this.numNodes; j++) {
                this.centerMoments[i][j] = v[i][j] + directedParams[i][this.parents[i].length - 1] *
                                                     directedParams[j][this.parents[j].length - 1];
                this.centerMoments[j][i] = this.centerMoments[i][j];
            }
            this.eNodes[i] = 0.;
            for (int j = 0; j < this.numNodes; j++)
                this.eNodes[i] += inv[i][j] * directedParams[j][this.parents[j].length - 1];
        }
        this.eeNodes = MatrixUtils.product(inv, MatrixUtils.product(this.centerMoments, MatrixUtils.transpose(inv)));
    }

    /**
     * Gets the matrix parameter for the G-Inverse Wishart posterior distribution using the current parameters
     * for directed edges.
     */

    protected double[][] getImpliedGInvWishartParam(double directedParams[][]) {
        for (int i = 0; i < this.numNodes; i++)
            for (int j = 0; j < this.numNodes; j++)
                this.iMinusB[i][j] = 0.;
        for (int i = 0; i < this.numNodes; i++)
            this.iMinusB[i][i] = 1.;
        for (int i = 0; i < this.numNodes; i++)
            for (int p = 0; p < this.parents[i].length - 1; p++)
                this.iMinusB[i][this.parents[i][p]] -= directedParams[i][p];
        double inv[][] = MatrixUtils.inverse(this.iMinusB);
        double mParam[] = new double[this.numNodes];
        for (int i = 0; i < this.numNodes; i++)
             mParam[i] = directedParams[i][this.parents[i].length - 1];
        double mean[] = MatrixUtils.product(inv, mParam);
        for (int i = 0; i < this.numNodes; i++)
            for (int j = 0; j < this.numNodes; j++)
                this.yCentered[i][j] = this.eeData[i][j] - mean[j] * this.eData[i] - mean[i] * this.eData[j] +
                                       this.sampleSize * mean[i] * mean[j];
        double prior[][] = MatrixUtils.product(this.iMinusB,
                                MatrixUtils.product(this.yCentered, MatrixUtils.transpose(this.iMinusB)));
        for (int i = 0; i < this.numNodes; i++)
            for (int j = 0; j < this.numNodes; j++)
                prior[i][j] += this.vPrior[i][j];
        return prior;
    }

    /**
     * Main step 1: do the actual sampling of the covariance matrix of error terms.
     */

    //The importance sampler. Weights are returned in log scale

    protected List<double[][]> sampleBidirectedMatrixImportance(Random r, double df, double paramMatrix[][]) {
        double weight = 0.;
        int    distOffset = 0;
        double matrix[][] = new double[this.numNodes][this.numNodes];
        double invOutputMatrix[][] = new double[this.numNodes][this.numNodes];

        for (int d = 0; d < this.numDistricts; d++) {
            double q = paramMatrix[distOffset][distOffset];
            matrix[distOffset][distOffset] = RbasProbUtils.inverseGammaRand(0.5 * df, 0.5 * q, r);
            weight += ProbUtils.lngamma(0.5 * df) - 0.5 * df * Math.log(0.5 * paramMatrix[distOffset][distOffset]);
            for (int i = 1; i < this.districts[d].length; i++) {
                //Bayesian regression of y_i on y_0, y_1, ..., y_{i - 1}
                double sigmaI[][] = new double[i][i];
                for (int v1 = 0; v1 < i; v1++)
                    for (int v2 = 0; v2 < i; v2++)
                         sigmaI[v1][v2] = paramMatrix[distOffset + v1][distOffset + v2];
                double invSigmaI[][] = MatrixUtils.inverse(sigmaI);
                double sigmaIPlusOneI[] = new double[i];
                for (int v1 = 0; v1 < i; v1++)
                    sigmaIPlusOneI[v1] = paramMatrix[distOffset + i][distOffset + v1];
                q = paramMatrix[distOffset + i][distOffset + i] - MatrixUtils.innerProduct(sigmaIPlusOneI,
                                                                  MatrixUtils.product(invSigmaI, sigmaIPlusOneI));
                double gamma;
                double Ri[][] = null;
                double beta[] = new double[i];
                double bStar[] = MatrixUtils.product(sigmaIPlusOneI, invSigmaI); //"default" mean of spouse/nonspouse b

                weight += 0.5 * this.numPreSpouses[d][i] * Math.log(2. * Math.PI)
                        + ProbUtils.lngamma(0.5 * (df + i + this.numNotPreSpouses[d][i]));

                if (this.numPreSpouses[d][i] == 0) {           //no spouses case
                    for (int v = 0; v < i; v++)
                        beta[v] = 0.;
                    double gammaFactor = MatrixUtils.innerProduct(bStar, MatrixUtils.product(sigmaI, bStar));
                    gamma = RbasProbUtils.inverseGammaRand(0.5 * (df + i + this.numNotPreSpouses[d][i]),
                            0.5 * (q + gammaFactor), r);
                    weight -= 0.5 * (df + i + this.numNotSpouses[d][i]) * Math.log(0.5 * (q + gammaFactor));
                    double nspNspCov[][] = new double[this.numNotPreSpouses[d][i]][this.numNotPreSpouses[d][i]];
                    for (int v1 = 0; v1 < this.numNotPreSpouses[d][i]; v1++)
                        for (int v2 = 0; v2 < this.numNotPreSpouses[d][i]; v2++)
                            nspNspCov[v1][v2] = matrix[distOffset + this.notPreSpouses[d][i][v1]][distOffset + this.notPreSpouses[d][i][v2]];
                    Ri = nspNspCov;
                } else if (this.numNotPreSpouses[d][i] == 0) { //no nonspouses case
                    gamma = RbasProbUtils.inverseGammaRand(0.5 * (df + i), 0.5 * q, r);
                    double normals[] = new double[i];
                    for (int v = 0; v < i; v++)
                         normals[v] = r.nextGaussian();
                    double cholesky[][] = MatrixUtils.cholesky(invSigmaI);
                    beta = MatrixUtils.product(cholesky, normals);
                    for (int v = 0; v < i; v++) {
                        beta[v] *= Math.sqrt(gamma);
                        beta[v] += bStar[v];
                    }
                    weight += 0.5 * Math.log(MatrixUtils.determinant(invSigmaI))
                            - 0.5 * (df + i + this.numNotPreSpouses[d][i]) * Math.log(0.5 * q);
                } else {                                        // spouses and nonspouses case
                    //Get the matrix that defines the linear relationship between betaNonspouses and betaSpouses
                    double spNspCov[][] = new double[this.numPreSpouses[d][i]][this.numNotPreSpouses[d][i]];
                    for (int v1 = 0; v1 < this.numPreSpouses[d][i]; v1++)
                        for (int v2 = 0; v2 < this.numNotPreSpouses[d][i]; v2++)
                            spNspCov[v1][v2] = matrix[distOffset + this.preSpouses[d][i][v1]][distOffset + this.notPreSpouses[d][i][v2]];
                    double nspNspCov[][] = new double[this.numNotPreSpouses[d][i]][this.numNotPreSpouses[d][i]];
                    for (int v1 = 0; v1 < this.numNotPreSpouses[d][i]; v1++)
                        for (int v2 = 0; v2 < this.numNotPreSpouses[d][i]; v2++)
                            nspNspCov[v1][v2] = matrix[distOffset + this.notPreSpouses[d][i][v1]][distOffset + this.notPreSpouses[d][i][v2]];
                    Ri = nspNspCov;
                    double A[][] = MatrixUtils.product(spNspCov, MatrixUtils.inverse(nspNspCov));
                    double AT[][] = MatrixUtils.transpose(A);

                    //Get the new mean and covariance for betaSpouses
                    double msp[] = new double[this.numPreSpouses[d][i]];
                    double mnsp[] = new double[this.numNotPreSpouses[d][i]];
                    double spSpU[][] = new double[this.numPreSpouses[d][i]][this.numPreSpouses[d][i]];
                    double spNspU[][] = new double[this.numPreSpouses[d][i]][this.numNotPreSpouses[d][i]];
                    double nspNspU[][] = new double[this.numNotPreSpouses[d][i]][this.numNotPreSpouses[d][i]];
                    for (int v1 = 0; v1 < this.numPreSpouses[d][i]; v1++) {
                        int index1 = this.preSpouses[d][i][v1];
                        msp[v1] = bStar[index1];
                        for (int v2 = 0; v2 < this.numPreSpouses[d][i]; v2++) {
                            int index2 = this.preSpouses[d][i][v2];
                            spSpU[v1][v2] = sigmaI[index1][index2];
                        }
                        for (int v2 = 0; v2 < this.numNotPreSpouses[d][i]; v2++) {
                            int index2 = this.notPreSpouses[d][i][v2];
                            spNspU[v1][v2] = sigmaI[index1][index2];
                        }
                    }
                    for (int v1 = 0; v1 < this.numNotPreSpouses[d][i]; v1++) {
                        int index1 = this.notPreSpouses[d][i][v1];
                        mnsp[v1] = bStar[index1];
                        for (int v2 = 0; v2 < this.numNotPreSpouses[d][i]; v2++) {
                            int index2 = this.notPreSpouses[d][i][v2];
                            nspNspU[v1][v2] = sigmaI[index1][index2];
                        }
                    }
                    double transSpNsp[][] = MatrixUtils.product(spNspU, AT);
                    double transNspNsp[][] = MatrixUtils.product(A, nspNspU);
                    double newInvCov[][] = MatrixUtils.product(transNspNsp, AT);
                    for (int v1 = 0; v1 < this.numPreSpouses[d][i]; v1++)
                        for (int v2 = 0; v2 < this.numPreSpouses[d][i]; v2++)
                            newInvCov[v1][v2] += spSpU[v1][v2] - transSpNsp[v1][v2] - transSpNsp[v2][v1];
                    double newCov[][] = MatrixUtils.inverse(newInvCov);
                    double newMean[] = MatrixUtils.product(newCov, MatrixUtils.sum(
                                          MatrixUtils.product(MatrixUtils.subtract(spSpU, MatrixUtils.transpose(transSpNsp)), msp),
                                          MatrixUtils.product(MatrixUtils.subtract(spNspU, transNspNsp), mnsp)));

                    //Compute new gamma distribution parameter
                    double gammaFactor = -MatrixUtils.innerProduct(newMean, MatrixUtils.product(newInvCov, newMean))
                                       +  MatrixUtils.innerProduct(msp, MatrixUtils.product(spSpU, msp))
                                       +  2 * MatrixUtils.innerProduct(msp, MatrixUtils.product(spNspU, mnsp))
                                       +  MatrixUtils.innerProduct(mnsp, MatrixUtils.product(nspNspU, mnsp));

                    gamma = RbasProbUtils.inverseGammaRand(0.5 * (df + i + this.numNotPreSpouses[d][i]),
                            0.5 * (q + gammaFactor), r);
                    weight += 0.5 * Math.log(MatrixUtils.determinant(newCov))
                            - 0.5 * (df + i + this.numNotSpouses[d][i]) * Math.log(0.5 * (q + gammaFactor));

                    //Finally sample betaSpouses and calculate betaNonSpouses
                    for (int v1 = 0; v1 < this.numPreSpouses[d][i]; v1++)
                        for (int v2 = 0; v2 < this.numPreSpouses[d][i]; v2++)
                            newCov[v1][v2] *= gamma;
                    double choleskyNewCov[][] = MatrixUtils.cholesky(newCov);
                    double normals[] = new double[this.numPreSpouses[d][i]];
                    for (int v = 0; v < this.numPreSpouses[d][i]; v++)
                         normals[v] = r.nextGaussian();
                    double betaSpouses[] = MatrixUtils.product(choleskyNewCov, normals);
                    for (int v = 0; v < this.numPreSpouses[d][i]; v++)
                        betaSpouses[v] += newMean[v];
                    double betaNSpouses[] = MatrixUtils.product(betaSpouses, A);
                    int countS = 0, countNS = 0;
                    for (int v = 0; v < i; v++) {
                        int index = distOffset + v;
                        if (!this.bidirectedMatrix[distOffset + i][index])
                            beta[v] = -betaNSpouses[countNS++];
                        else
                            beta[v] = betaSpouses[countS++];
                    }
                }

                //Fill in the new elements of the covariance matrix.
                for (int v = 0; v < i; v++) {
                    matrix[distOffset + i][distOffset + v] = 0.;
                    for (int v2 = 0; v2 < i; v2++)
                        matrix[distOffset + i][distOffset + v] += beta[v2] * matrix[distOffset + v][distOffset + v2];
                    matrix[distOffset + v][distOffset + i] = matrix[distOffset + i][distOffset + v];
                }
                for (int v1 = 0; v1 < i; v1++)
                    sigmaIPlusOneI[v1] = matrix[distOffset + i][distOffset + v1];
                matrix[distOffset + i][distOffset + i] = gamma + MatrixUtils.innerProduct(beta, sigmaIPlusOneI);

                //Finalize weight with Jacobian, if necessary
                if (this.numNotPreSpouses[d][i] > 0)
                    weight -= Math.log(MatrixUtils.determinant(Ri));
            }
            if (Double.isNaN(matrix[distOffset][distOffset]))
                throw new RuntimeException("NaN values obtained while sampling. Quitting now.");
            if (this.districts[d].length > 1) {
                for (int i = 0; i < this.districts[d].length; i++)
                    for (int j = 0; j < this.districts[d].length; j++)
                        this.invBuffer[d][i][j] = matrix[distOffset + i][distOffset + j];
                double inverse[][] = MatrixUtils.inverse(invBuffer[d]);
                for (int i = 0; i < this.districts[d].length; i++)
                    for (int j = 0; j < this.districts[d].length; j++)
                        invOutputMatrix[distOffset + i][distOffset + j] = inverse[i][j];
            } else
               invOutputMatrix[distOffset][distOffset] = 1. / matrix[distOffset][distOffset];
            distOffset += this.districts[d].length;
        }

        List<double[][]> output = new ArrayList<double[][]>(); double w[][] = new double[1][1]; w[0][0] = weight;
        output.add(matrix); output.add(invOutputMatrix); output.add(w);
        return output;
    }

    protected List<double[][]> sampleBidirectedMatrixGibbs(Random r, double df, double paramMatrix[][], double previous[][]) {
        int    distOffset = 0;
        double matrix[][] = new double[this.numNodes][this.numNodes];
        for (int i = 0; i < this.numNodes; i++)
            System.arraycopy(previous[i], 0, matrix[i], 0, this.numNodes);

        double invOutputMatrix[][] = new double[this.numNodes][this.numNodes];
        for (int d = 0; d < this.numDistricts; d++) {
            double sigmaI[][] = new double[this.districts[d].length - 1][this.districts[d].length - 1];
            double beta[] = new double[this.districts[d].length - 1];
            for (int i = 0; i < this.districts[d].length; i++) {
                //Regression of y_i on y_0, y_1, ..., y_{i - 1}
                for (int v1 = 0; v1 < this.districts[d].length - 1; v1++) {
                    int index1 = distOffset + v1;
                    if (v1 >= i)
                        index1++;
                    for (int v2 = 0; v2 < this.districts[d].length - 1; v2++) {
                        int index2 = distOffset + v2;
                        if (v2 >= i)
                            index2++;
                        sigmaI[v1][v2] = paramMatrix[index1][index2];
                    }
                }
                double invSigmaI[][] = MatrixUtils.inverse(sigmaI);
                double sigmaIPlusOneI[] = new double[this.districts[d].length - 1];
                for (int v1 = 0; v1 < this.districts[d].length - 1; v1++) {
                    int index = distOffset + v1;
                    if (v1 >= i)
                        index++;
                    sigmaIPlusOneI[v1] = paramMatrix[distOffset + i][index];
                }
                double bStar[] = MatrixUtils.product(sigmaIPlusOneI, invSigmaI); //mean of spouse/nonspouse b
                double q = paramMatrix[distOffset + i][distOffset + i] - MatrixUtils.innerProduct(bStar, sigmaIPlusOneI);
                double gamma;


                if (this.numSpouses[d][i] == 0) {           //no spouses case
                    double gammaFactor = MatrixUtils.innerProduct(bStar, MatrixUtils.product(sigmaI, bStar));
                    for (int v = 0; v < this.districts[d].length - 1; v++)
                        beta[v] = 0.;
                    gamma = RbasProbUtils.inverseGammaRand(
                            0.5 * (df + (this.districts[d].length - 1) + this.numNotSpouses[d][i]),
                            0.5 * (q + gammaFactor), r);
                } else if (this.numNotSpouses[d][i] == 0) { //no nonspouses case
                    gamma = RbasProbUtils.inverseGammaRand(
                            0.5 * (df + (this.districts[d].length - 1) + this.numNotSpouses[d][i]), 0.5 * q, r);
                    double normals[] = new double[this.districts[d].length - 1];
                    for (int v = 0; v < this.districts[d].length - 1; v++)
                         normals[v] = r.nextGaussian();
                    double cholesky[][] = MatrixUtils.cholesky(invSigmaI);
                    beta = MatrixUtils.product(cholesky, normals);
                    for (int v = 0; v < this.districts[d].length - 1; v++) {
                        beta[v] *= Math.sqrt(gamma);
                        beta[v] += bStar[v];
                    }
                } else {                                        // spouses and nonspouses case
                    //Get the matrix that defines the linear relationship between betaNonspouses and betaSpouses
                    double spNspCov[][] = new double[this.numSpouses[d][i]][this.numNotSpouses[d][i]];
                    for (int v1 = 0; v1 < this.numSpouses[d][i]; v1++)
                        for (int v2 = 0; v2 < this.numNotSpouses[d][i]; v2++)
                            spNspCov[v1][v2] = matrix[distOffset + this.spouses[d][i][v1]][distOffset + this.notSpouses[d][i][v2]];
                    double nspNspCov[][] = new double[this.numNotSpouses[d][i]][this.numNotSpouses[d][i]];
                    for (int v1 = 0; v1 < this.numNotSpouses[d][i]; v1++)
                        for (int v2 = 0; v2 < this.numNotSpouses[d][i]; v2++)
                            nspNspCov[v1][v2] = matrix[distOffset + this.notSpouses[d][i][v1]][distOffset + this.notSpouses[d][i][v2]];
                    double A[][] = MatrixUtils.product(spNspCov, MatrixUtils.inverse(nspNspCov));
                    double AT[][] = MatrixUtils.transpose(A);

                    //Get the new mean and covariance for betaSpouses
                    double msp[] = new double[this.numSpouses[d][i]], mnsp[] = new double[this.numNotSpouses[d][i]];
                    double spSpU[][] = new double[this.numSpouses[d][i]][this.numSpouses[d][i]];
                    double spNspU[][] = new double[this.numSpouses[d][i]][this.numNotSpouses[d][i]];
                    double nspNspU[][] = new double[this.numNotSpouses[d][i]][this.numNotSpouses[d][i]];
                    for (int v1 = 0; v1 < this.numSpouses[d][i]; v1++) {
                        int index1 = this.spouses[d][i][v1]; if (index1 > i) index1--;
                        msp[v1] = bStar[index1];
                        for (int v2 = 0; v2 < this.numSpouses[d][i]; v2++) {
                            int index2 = this.spouses[d][i][v2]; if (index2 > i) index2--;
                            spSpU[v1][v2] = sigmaI[index1][index2];
                        }
                        for (int v2 = 0; v2 < this.numNotSpouses[d][i]; v2++) {
                            int index2 = this.notSpouses[d][i][v2]; if (index2 > i) index2--;
                            spNspU[v1][v2] = sigmaI[index1][index2];
                        }
                    }
                    for (int v1 = 0; v1 < this.numNotSpouses[d][i]; v1++) {
                        int index1 = this.notSpouses[d][i][v1]; if (index1 > i) index1--;
                        mnsp[v1] = bStar[index1];
                        for (int v2 = 0; v2 < this.numNotSpouses[d][i]; v2++) {
                            int index2 = this.notSpouses[d][i][v2]; if (index2 > i) index2--;
                            nspNspU[v1][v2] = sigmaI[index1][index2];
                        }
                    }
                    double transSpNsp[][] = MatrixUtils.product(spNspU, AT);
                    double transNspNsp[][] = MatrixUtils.product(A, nspNspU);
                    double newInvCov[][] = MatrixUtils.product(transNspNsp, AT);
                    for (int v1 = 0; v1 < this.numSpouses[d][i]; v1++)
                        for (int v2 = 0; v2 < this.numSpouses[d][i]; v2++)
                            newInvCov[v1][v2] += spSpU[v1][v2] - transSpNsp[v1][v2] - transSpNsp[v2][v1];
                    double newCov[][] = MatrixUtils.inverse(newInvCov);
                    double newMean[] = MatrixUtils.product(newCov, MatrixUtils.sum(
                                          MatrixUtils.product(MatrixUtils.subtract(spSpU, MatrixUtils.transpose(transSpNsp)), msp),
                                          MatrixUtils.product(MatrixUtils.subtract(spNspU, transNspNsp), mnsp)));

                    //Compute new gamma distribution parameter
                    //double gammaFactor = -MatrixUtils.innerProduct(newMean, MatrixUtils.product(newInvCov, newMean))
                    //                   +  MatrixUtils.innerProduct(msp, MatrixUtils.product(spSpU, msp))
                    //                   +  2 * MatrixUtils.innerProduct(msp, MatrixUtils.product(spNspU, mnsp))
                    //                   +  MatrixUtils.innerProduct(mnsp, MatrixUtils.product(nspNspU, mnsp));
                    double gammaFactor = -MatrixUtils.innerProduct(newMean, MatrixUtils.product(newInvCov, newMean))
                                       +  MatrixUtils.innerProduct(msp, MatrixUtils.product(spSpU, msp))
                                       +  2 * MatrixUtils.innerProduct(msp, MatrixUtils.product(spNspU, mnsp))
                                       +  MatrixUtils.innerProduct(mnsp, MatrixUtils.product(nspNspU, mnsp));
                    gamma = RbasProbUtils.inverseGammaRand(0.5 * (df + (this.districts[d].length - 1) + this.numNotSpouses[d][i]),
                            0.5 * (q + gammaFactor), r);

                    for (int v1 = 0; v1 < this.numSpouses[d][i]; v1++)
                        for (int v2 = 0; v2 < this.numSpouses[d][i]; v2++)
                            newCov[v1][v2] *= gamma;
                    double choleskyNewCov[][] = MatrixUtils.cholesky(newCov);

                    //Finally sample betaSpouses and calculate betaNonSpouses
                    double normals[] = new double[this.numSpouses[d][i]];
                    for (int v = 0; v < this.numSpouses[d][i]; v++)
                         normals[v] = r.nextGaussian();
                    double betaSpouses[] = MatrixUtils.product(choleskyNewCov, normals);
                    for (int v = 0; v < this.numSpouses[d][i]; v++)
                        betaSpouses[v] += newMean[v];
                    double betaNSpouses[] = MatrixUtils.product(betaSpouses, A);
                    int countS = 0, countNS = 0;
                    for (int v = 0; v < this.districts[d].length - 1; v++) {
                        int index = distOffset + v;
                        if (v >= i)
                            index++;
                        if (!this.bidirectedMatrix[distOffset + i][index])
                            beta[v] = -betaNSpouses[countNS++];
                        else
                            beta[v] = betaSpouses[countS++];
                    }
                }
                //Fill in the new elements of the covariance matrix.
                for (int v = 0; v < this.districts[d].length - 1; v++) {
                    int index1 = distOffset + v;
                    if (v >= i)
                        index1++;
                    matrix[distOffset + i][index1] = 0.;
                    for (int v2 = 0; v2 < this.districts[d].length - 1; v2++) {
                        int index2 = distOffset + v2;
                        if (v2 >= i)
                            index2++;
                        matrix[distOffset + i][index1] += beta[v2] * matrix[index1][index2];
                    }
                    matrix[index1][distOffset + i] = matrix[distOffset + i][index1];
                }
                for (int v1 = 0; v1 < this.districts[d].length - 1; v1++) {
                    int index1 = distOffset + v1;
                    if (v1 >= i)
                        index1++;
                    sigmaIPlusOneI[v1] = matrix[distOffset + i][index1];
                }
                matrix[distOffset + i][distOffset + i] = gamma + MatrixUtils.innerProduct(beta, sigmaIPlusOneI);
            }
            if (Double.isNaN(matrix[distOffset][distOffset]))
                throw new RuntimeException("NaN values obtained while sampling. Quitting now.");
            if (this.districts[d].length > 1) {
                for (int i = 0; i < this.districts[d].length; i++)
                    for (int j = 0; j < this.districts[d].length; j++)
                        this.invBuffer[d][i][j] = matrix[distOffset + i][distOffset + j];
                double inverse[][] = MatrixUtils.inverse(this.invBuffer[d]);
                for (int i = 0; i < this.districts[d].length; i++)
                    for (int j = 0; j < this.districts[d].length; j++)
                        invOutputMatrix[distOffset + i][distOffset + j] = inverse[i][j];
            } else
               invOutputMatrix[distOffset][distOffset] = 1. / matrix[distOffset][distOffset];
            distOffset += this.districts[d].length;
        }

        List<double[][]> output = new ArrayList<double[][]>(); double w[][] = new double[1][1]; w[0][0] = 1.;
        output.add(matrix); output.add(invOutputMatrix); output.add(w);
        return output;
    }

    /**
     * Main step 2 (I): do the actual sampling of the parameters for directed edges. Sample each coefficient conditioned
     * on all others. Simpler to implement than the block-sample, since fixed coefficients are just ignored.
     * Arguably not as useful, since mixing times are prone to be longer. The following code is more for educational
     * purposes.
     */

    protected double[][] sampleDirectedMatrix(Random r, double errorMatrix[][],double invErrMatrix[][],
                                              double previousParams[][], double data[][]) {
        double directedMatrix[][] = new double[this.numNodes][];
        for (int i = 0; i < this.numNodes; i++) {
            directedMatrix[i] = new double[this.parents[i].length];
            System.arraycopy(previousParams[i], 0, directedMatrix[i], 0, directedMatrix[i].length);
        }
        for (int i = 0; i < this.numNodes; i++) {
            for (int pj = 0; pj < directedMatrix[i].length; pj++) {
                if (this.fixedParams[i][pj])
                    continue;
                int j = this.parents[i][pj];
                double variance = 0.;
                for (int s = 0; s < data.length; s++)
                    variance += data[s][j] * data[s][j];
                variance = 1. / (invErrMatrix[i][i] * variance + 1. / this.bPrior[i][pj][1]);
                double mean = this.bPrior[i][pj][0] / this.bPrior[i][pj][1];
                for (int s = 0; s < data.length; s++) {
                    double aux = 0.;
                    //Other parents of the same node i
                    double aux2 = data[s][i];
                    for (int pOthers = 0; pOthers < directedMatrix[i].length; pOthers++) {
                        if (pOthers == pj)
                            continue;
                        aux2 -= directedMatrix[i][pOthers] * data[s][this.parents[i][pOthers]];
                    }
                    aux += invErrMatrix[i][i] * aux2;
                    //Other nodes in the same district and their parents
                    for (int pp = 0; pp < this.districtNeighbors[i].length; pp++) {
                        int p = this.districtNeighbors[i][pp];
                        aux2 = data[s][p];
                        for (int pq = 0; pq < this.parents[p].length; pq++) {
                            int q = this.parents[p][pq];
                            aux2 -= directedMatrix[p][pq] * data[s][q];
                        }
                        aux2 *= invErrMatrix[i][p];
                        aux += aux2;
                    }
                    mean += data[s][j] * aux;
                }
                mean *= variance;
                directedMatrix[i][pj] = r.nextGaussian() * Math.sqrt(variance) + mean;
            }
        }

        return directedMatrix;
    }

    /**
     * Main step 2 (II): do the actual sampling of the parameters for directed edges. Factorize elements according
     * to the districts of the graph and sample all coefficients in block. One has to consider condition on the
     * fixed coefficients, though. This is a much more complicated implementation, since the matrix of coefficients has
     * to be mapped into a vector while ignoring the fixed coefficients.
     */

    protected double[][] sampleDirectedMatrixBlock(Random r, double errorMatrix[][],double invErrMatrix[][],
                                                   double previousParams[][]) {
        double directedMatrix[][] = new double[this.numNodes][];
        for (int v = 0; v < this.numNodes; v++) {
            directedMatrix[v] = new double[this.parents[v].length];
            //Copy fixed parameter values
            for (int p = 0; p < this.parents[v].length; p++)
                if (this.fixedParams[v][p])
                   directedMatrix[v][p] = previousParams[v][p];
        }
        for (int d = 0; d < this.numDistricts; d++) {
            //First, compute first and second moments of the coefficients
            for (int i = 0; i < this.indexCoeff[d].length; i++) {
                int vi = this.indexCoeff[d][i][0], vip = this.indexCoeff[d][i][1];
                //Second moments
                for (int j = i; j < this.indexCoeff[d].length; j++) {
                    int vj = this.indexCoeff[d][j][0], vjp = this.indexCoeff[d][j][1];
                    this.eeCoeff[d][i][j] = this.eeCoeff[d][j][i] =
                            this.eeData[this.parents[vi][vip]][this.parents[vj][vjp]] *
                                            invErrMatrix[vi][vj];
                }
                //First moments
                this.eCoeff[d][i] = this.priorMeanCovCoeff[d][i];
                for (int v = 0; v < this.districts[d].length; v++) {
                    int vv = this.districts[d][v];
                    this.eCoeff[d][i] += invErrMatrix[vi][vv] * this.eeData[this.parents[vi][vip]][vv];
                    for (int p = 0; p < this.parents[vv].length; p++)
                        if (this.fixedParams[vv][p]) //Take care of fixed params
                            this.eCoeff[d][i] -= invErrMatrix[vi][vv] *
                                                 previousParams[vv][p] *
                                                 this.eeData[this.parents[vi][vip]][this.parents[vv][p]];
                }

            }
            //Covariance & mean
            for (int i = 0; i < this.eCoeff[d].length; i++)
                this.eeCoeff[d][i][i] += this.priorCovCoeff[d][i];
            double covCoeff[][] = MatrixUtils.inverse(this.eeCoeff[d]);
            double meanCoeff[] = MatrixUtils.product(covCoeff, this.eCoeff[d]);
            //Finally, sample values
            double cholesky[][] = MatrixUtils.cholesky(covCoeff);
            for (int i = 0; i < this.cBuffer1[d].length; i++)
                this.cBuffer1[d][i] = r.nextGaussian();
            for (int i = 0; i < this.cBuffer2[d].length; i++) {
                this.cBuffer2[d][i] = meanCoeff[i];
                for (int j = 0; j < this.cBuffer1[d].length; j++)
                    this.cBuffer2[d][i] += cholesky[i][j] * this.cBuffer1[d][j];
            }
            //Map back to output matrix
            for (int i = 0; i < this.indexCoeff[d].length; i++) {
                int vi = this.indexCoeff[d][i][0], vip = this.indexCoeff[d][i][1];
                directedMatrix[vi][vip] = this.cBuffer2[d][i];
                if (Double.isInfinite(directedMatrix[vi][vip]))
                    logger.finest("!" + d);
            }
        }
        return directedMatrix;
    }

    /**
     * Calculates the predictive log-likelihood for the new data.
     */

    public double predictiveLoglikelihood(List inputSamples, RectangularDataSet data, int burnIn, int simulationSize, int stepSize) {
        List samples;
        if (inputSamples == null)
            samples = doSample(simulationSize);
        else
            samples = inputSamples;
        double iMinusB[][] = new double[this.numNodes][this.numNodes];
        double intercept[] = new double[this.numNodes];
        double meanObs[] = new double[this.numObserved];
        double covObs[][] = new double[this.numObserved][this.numObserved];
        int    totalSamples = 0;
        double dummyRawData[][] = data.getDoubleData().toArray();
        List<String> testVariableNames = data.getVariableNames();
        int    testSampleSize = data.getNumRows();
        double testRawData[][] = new double[testSampleSize][this.numObserved]; int columnCount = 0;
        for (int i = 0; i < this.numNodes; i++) {
            if (this.isLatent[i])
                continue;
            String varName = this.graph.getNodes().get(i).getName();
            for (int j = 0; j < this.numObserved; j++) {
                if (!testVariableNames.get(j).equals(varName))
                    continue;
                for (int d = 0; d < testSampleSize; d++)
                    testRawData[d][columnCount] = dummyRawData[d][j];
            }
            columnCount++;
        }
        double dataProduct[][] = new double[this.numObserved][this.numObserved];
        double likelihoods[] = new double[testSampleSize];
        double testPoint[] = new double[this.numObserved];
        double allBidirected[][][] = (double[][][]) samples.get(0);
        double allDirected[][][] = (double[][][]) samples.get(2);
        for (int s = burnIn; s < simulationSize; s += stepSize) {
            double errSample[][] = unpackCovariance(allBidirected[s]);
            double coeffSample[][] = allDirected[s];
            for (int i = 0; i < this.numNodes; i++)
                for (int j = 0; j < this.numNodes; j++)
                    iMinusB[i][j] = 0.;
            for (int i = 0; i < this.numNodes; i++)
                iMinusB[i][i] = 1.;
            for (int i = 0; i < this.numNodes; i++) {
                for (int j = 0; j < this.parents[i].length - 1; j++)
                    iMinusB[i][this.parents[i][j]] -= coeffSample[i][j];
                intercept[i] = coeffSample[i][this.parents[i].length - 1];
            }
            double invIMinusB[][] = MatrixUtils.inverse(iMinusB);
            double mean[] = MatrixUtils.product(invIMinusB, intercept);
            double implCov[][] = MatrixUtils.product(invIMinusB, MatrixUtils.product(errSample, MatrixUtils.transpose(invIMinusB)));
            int columnCount1 = 0;
            for (int i = 0; i < this.numNodes; i++) {
                if (this.isLatent[i])
                    continue;
                meanObs[columnCount1] = mean[i];
                int columnCount2 = 0;
                for (int j = 0; j < this.numNodes; j++) {
                    if (this.isLatent[j])
                        continue;
                    covObs[columnCount1][columnCount2++] = implCov[i][j];
                }
                columnCount1++;
            }

            double invCovObs[][] = MatrixUtils.inverse(covObs);
            double sqrtDetCovObs = Math.sqrt(MatrixUtils.determinant(covObs));
            for (int d = 0; d < testSampleSize; d++) {
                for (int j = 0; j < this.numObserved; j++)
                    testPoint[j] = testRawData[d][j] - mean[j];
                for (int i = 0; i < this.numObserved; i++)
                    for (int j = 0; j < this.numObserved; j++)
                        dataProduct[i][j] = testPoint[i] * testPoint[j];
                likelihoods[d] += Math.exp(-0.5 * MatrixUtils.trace(MatrixUtils.product(dataProduct, invCovObs))) /
                                  sqrtDetCovObs;
            }
            totalSamples++;
        }

        double testLikelihood = testSampleSize * (-0.5 * this.numObserved * Math.log(2. * Math.PI));
        for (int d = 0; d < testSampleSize; d++)
            testLikelihood += Math.log(likelihoods[d] / totalSamples);
        return testLikelihood;
    }


    /******************************************************************************************************************
     * OTHER METHODS
     */

    public double[][] packCovariance(double C[][]) {
        double output[][] = new double[this.numNodes][];
        int distOffset = 0;
        for (int d = 0; d < this.districts.length; d++) {
            for (int i = 0; i < this.districts[d].length; i++) {
                int v = this.districts[d][i];
                output[v] = new double[this.numPreSpouses[d][i] + 1];
                for (int s = 0; s < this.numPreSpouses[d][i]; s++)
                    output[v][s] = C[v][distOffset + this.preSpouses[d][i][s]];
                output[v][this.numPreSpouses[d][i]] = C[v][v];
            }
            distOffset += this.districts[d].length;
        }
        return output;
    }

    public double[][] unpackCovariance(double C[][]) {
        double output[][] = new double[this.numNodes][this.numNodes];
        int distOffset = 0;
        for (int d = 0; d < this.districts.length; d++) {
            for (int i = 0; i < this.districts[d].length; i++) {
                int v = this.districts[d][i];
                for (int s = 0; s < this.numPreSpouses[d][i]; s++) {
                    output[v][distOffset + this.preSpouses[d][i][s]] = C[v][s];
                    output[distOffset + this.preSpouses[d][i][s]][v] = C[v][s];
                }
                output[v][v] = C[v][this.numPreSpouses[d][i]];
            }
            distOffset += this.districts[d].length;
        }
        return output;
    }

    public Graph getGraph() {
        return this.graph;
    }

    public RectangularDataSet getData() {
        return this.train;
    }

    public void setBlockLatentSampling(boolean b) {
        this.blockLatentSampling = b;
    }

    public boolean getBlockLatentSampling() {
        return this.blockLatentSampling;
    }

    public RectangularDataSet getReorderedData() {
        return this.reorderedData;
    }

    public boolean getReturnLatents() {
        return this.returnLatents;
    }

    public int getNumObserved() {
        return this.numObserved;
    }

    public int getNumLatents() {
        return this.numLatents;
    }

    public int[] getObservedIndex() {
        return this.observedIndex;
    }

    public void setReturnLatents(boolean rl) {
        this.returnLatents = rl;
    }

    //This centers the data at its empirical mean if requested, and then assumes a zero-mean model.

    public List<List<double[][]>> bidirectedSampler(int simulationSize, boolean center, boolean priorOnly) {
        double df = this.vPriorDf;
        double k[][] = new double[this.numNodes][this.numNodes];
        for (int v1 = 0; v1 < this.numNodes; v1++)
            for (int v2 = 0; v2 < this.numNodes; v2++)
                k[v1][v2] = this.vPrior[v1][v2];
        if (!priorOnly) {
            df += this.sampleSize;
            if (center) {
                double mean[] = new double[this.numNodes];
                double centeredData[][] = new double[this.sampleSize][this.numNodes];
                for (int v1 = 0; v1 < this.numNodes; v1++)
                    for (int d = 0; d < this.sampleSize; d++)
                        mean[v1] += this.rawData[d][v1];
                for (int v1 = 0; v1 < this.numNodes; v1++)
                    mean[v1] /= this.sampleSize;
                for (int d = 0; d < this.sampleSize; d++)
                    for (int v1 = 0; v1 < this.numNodes; v1++)
                         centeredData[d][v1] = this.rawData[d][v1] - mean[v1];
                for (int d = 0; d < this.sampleSize; d++)
                    for (int v1 = 0; v1 < this.numNodes; v1++)
                        for (int v2 = 0; v2 < this.numNodes; v2++)
                            k[v1][v2] += centeredData[d][v1] * centeredData[d][v2];
            } else {
                for (int d = 0; d < this.sampleSize; d++)
                    for (int v1 = 0; v1 < this.numNodes; v1++)
                        for (int v2 = 0; v2 < this.numNodes; v2++)
                            k[v1][v2] += this.rawData[d][v1] * this.rawData[d][v2];
            }
        }
        Random r = new Random();
        if (!RbasGraphUtils.isBidirected(this.graph))
            throw new RuntimeException("Improper call to bidirectedMarginalLoglikelihood: graph is not bi-directed");
        if (this.numLatents > 0)
            throw new RuntimeException("Improper call to bidirectedMarginalLoglikelihood: graph contains latent variables");
        List<List<double[][]>> allSamples = new ArrayList<List<double[][]>>();
        for (int s = 0; s < simulationSize; s++) {
            logger.finest(" - Sample #" + s);
            List<double[][]> sample = sampleBidirectedMatrixImportance(r, df, k);
            allSamples.add(sample);
        }
        return allSamples;
    }

    public double bidirectedMarginalLoglikelihood(int simulationSize, boolean priorOnly) {
        List<List<double[][]>>  samples = bidirectedSampler(simulationSize, false, priorOnly);

        double max = -Double.MAX_VALUE;
        for (List<double[][]> sample : samples)
            if (sample.get(2)[0][0] > max)
                max = sample.get(2)[0][0];

        double Z = 0.;
        for (List<double[][]> sample : samples)
            Z += Math.exp(sample.get(2)[0][0] - max);

        return Math.log(Z / simulationSize) + max;
    }

    public double bidirectedMarginalLoglikelihood(int simulationSize) {
        return bidirectedMarginalLoglikelihood(simulationSize, false);
    }

    //Log-normalizing constant of a inverseWishart using the model parameters

    public double inverseWishartZ() {
        if (!RbasGraphUtils.isBidirected(this.graph))
            throw new RuntimeException("Improper call to inverseWishartZ: graph is not bi-directed");
        if (this.numLatents > 0)
            throw new RuntimeException("Improper call to inverseWishartZ: graph contains latent variables");
        double df = this.vPriorDf + this.sampleSize;
        double k[][] = new double[this.numNodes][this.numNodes];
        for (int v1 = 0; v1 < this.numNodes; v1++)
            for (int v2 = 0; v2 < this.numNodes; v2++)
                k[v1][v2] = this.vPrior[v1][v2];
        for (int d = 0; d < this.sampleSize; d++)
            for (int v1 = 0; v1 < this.numNodes; v1++)
                for (int v2 = 0; v2 < this.numNodes; v2++)
                    k[v1][v2] += this.rawData[d][v1] * this.rawData[d][v2];
        return inverseWishartZ(df, k);
    }

    public static double inverseWishartZ(double df, double U[][]) {
        int    k = U.length;
        double upsilon = df + k - 1;
        double Z = 0.5 * upsilon * k * Math.log(2) + 0.25 * k * (k - 1) * Math.log(Math.PI);
        for (int i = 1; i <= k; i++)
            Z += ProbUtils.lngamma(0.5 * (upsilon + 1 - i));
        Z -= 0.5 * upsilon * Math.log(MatrixUtils.determinant(U));
        return Z;
    }

    public static double inverseWishartZ2(double df, double U[][]) {
        int    k = U.length;
        double Z = ProbUtils.lngamma(0.5 * df) - 0.5 * df * Math.log(0.5 * U[0][0]);
        for (int i = 1; i < k; i++) {
            Z += 0.5 * i * Math.log(2 * Math.PI);
            double innerU[][] = new double[i][i];
            double Ui[] = new double[i];
            for (int j = 0; j < i; j++) {
                Ui[j] = U[i][j];
                for (int t = 0; t < i; t++)
                    innerU[j][t] = U[j][t];
            }
            double u_ii = U[i][i] - MatrixUtils.innerProduct(Ui, MatrixUtils.product(MatrixUtils.inverse(innerU), Ui));
            Z -= 0.5 * Math.log(MatrixUtils.determinant(innerU));
            Z += ProbUtils.lngamma(0.5 * (df + i));
            Z -= 0.5 * (df + i) * Math.log(0.5 * u_ii);
        }
        return Z;
    }

    public static void resetLoggerHandler(StreamHandler handle) {
        Handler handlers[] = logger.getHandlers();
        for (int i = 0; i < handlers.length; i++)
            logger.removeHandler(handlers[i]);
        logger.addHandler(handle);
    }

    public static void setLoggerLevel(int level) {
        switch (level) {
            case 1: logger.setLevel(Level.OFF); break;
            case 2: logger.setLevel(Level.FINE); break;
            case 3: logger.setLevel(Level.FINER); break;
            case 4: logger.setLevel(Level.FINEST); break;
        }
    }

}
