package rbas.dmg.sem;

import edu.cmu.tetrad.graph.*;
import edu.cmu.tetrad.data.RectangularDataSet;
import edu.cmu.tetrad.util.MatrixUtils;

import java.util.List;
import java.util.ArrayList;
import java.util.Random;
import java.util.Hashtable;

import rbas.util.RbasProbUtils;

/**
 * McmcProbitMixedGraph
 *
 * Implements a Markov Chain Monte Carlo sampler for a mixed graph model with continuous and discrete variables.
 * Discrete variables are binary or ordinal. The probit link function is used. Continuous variables are Gaussian
 * distributed.
 *
 * Cyclic graphs are not supported.
 *
 * @author Ricardo Silva
 * @version $Revision: 1.0 $ $Date: 2007/09/21$
 */


public class McmcProbitMixedGraph extends McmcGaussianMixedGraph {

    double  ordValues[][];  //possible values that the binary/ordinal variables can assume
    int     numCat[];       //0 for continuous variables; 2 for binary variables; 3, 4, etc. for ordinal variables
    boolean isDiscrete[];

    double thresholdPrior[][][]; //mean and variance parameters for the gaussian distribution that defines the
                                 //prior over thresholds
    boolean fixedThreshold[][]; //indicates which thresholds are fixed at the mean of the prior

    double rawDataStar[][]; //includes the underlying latent variables of the discrete variables. Continuous
                            //variables will also have an underlying latent that is just set to have the same
                            //variable as themselves

    double residuals[];

    public McmcProbitMixedGraph(Graph graph, RectangularDataSet train, double vPrior[][], double df,
                                double bPrior[][][], boolean fixedB[][], double thresholdPrior[][][],
                                boolean fixedThreshold[][], double ordValues[][]) {
        super(graph, train, vPrior, df, bPrior, fixedB);
        this.thresholdPrior = thresholdPrior;
        this.fixedThreshold = fixedThreshold;
        startDiscreteInformation(ordValues, graph);
    }

    /**
     * Notice that because of the reordering of nodes as done in rebuildOrderedGraph, we can't just make a direct
     * copy of the discrete data information given as arguments to the constructors. This is the reason why we build a
     * mapping of nodes in the original order to the new order.
     */

    protected void startDiscreteInformation(double originalOrdValues[][], Graph oldGraph) {
        List nodes = this.graph.getNodes();
        List originalNodes = oldGraph.getNodes();
        int map[] = new int[nodes.size()];
        for (int i = 0; i < originalNodes.size(); i++) {
            int pos = -1;
            for (int j = 0; j < nodes.size(); j++)
                if (nodes.get(j) == originalNodes.get(i)) {
                    pos = j;
                    break;
                }
            map[i] = pos;
        }
        this.numCat = new int[this.numNodes];
        this.isDiscrete = new boolean[this.numNodes];
        this.ordValues = new double[this.numNodes][];
        for (int i = 0; i < this.numNodes; i++) {
            this.ordValues[map[i]] = originalOrdValues[i];
            if (!this.isLatent[map[i]]) {
                this.numCat[map[i]] = this.ordValues[map[i]].length;
                this.isDiscrete[map[i]] = this.numCat[map[i]] > 0;
                this.fixedThreshold[map[i]] = this.fixedThreshold[i];
                this.thresholdPrior[map[i]] = this.thresholdPrior[i];
            } else {
                this.isDiscrete[map[i]] = false;
            }
        }
    }

    protected void initialize() {
        super.initialize();
        this.residuals = new double[this.numNodes];
        this.rawDataStar = new double[this.rawData.length][this.numNodes];
    }


    /**
     * Sampling methods
     */

    public List doSample(int number) {
        List samples = new ArrayList();
        Random r = new Random();
        double samplesBidirected[][][] = new double[number][][];
        double samplesDirected[][][]   = new double[number][][];
        double sampleThresholds[][][]  = new double[number][][];
        double samplesWeight[]         = new double[number];
        double bidirected[][]          = new double[this.numNodes][this.numNodes];
        double directed[][]            = new double[this.numNodes][];
        double thresholds[][]          = new double[this.numNodes][];
        double latents[][][]              = null;

        if (this.returnLatents)
            latents = new double[number][this.sampleSize][this.numLatents];

        //Initialization: sample connection parameters
        for (int i = 0; i < this.numNodes; i++) {
            directed[i] = new double[this.parents[i].length];
            for (int p = 0; p < this.parents[i].length; p++) {
                if (!this.fixedParams[i][p])
                    directed[i][p] = r.nextGaussian();
                else
                    directed[i][p] = this.bPrior[i][p][0];
            }
        }
        for (int i = 0; i < this.numNodes; i++)
            bidirected[i][i] = 1. + r.nextDouble();
        
        //Initialization: sample thresholds
        for (int i = 0; i < this.numNodes; i++)
            if (this.isDiscrete[i]) {
                thresholds[i] = new double[this.numCat[i] - 1];
                for (int v = 0; v < this.numCat[i] - 1; v++) {
                    if (!this.fixedThreshold[i][v])
                        thresholds[i][v] = this.ordValues[i][v] + 0.5 * (this.ordValues[i][v + 1] - this.ordValues[i][v]);
                    else
                        thresholds[i][v] = this.thresholdPrior[i][v][0];
                }
            }

        //Initialization: sample underlying variables
        for (int i = 0; i < this.numNodes; i++) {
            if (this.isLatent[i])
                for (int j = 0; j < this.sampleSize; j++)
                    this.rawData[j][i] = r.nextGaussian();
            if (this.isDiscrete[i]) {
                for (int j = 0; j < this.sampleSize; j++) {
                    double l = getOrdinalLowerBound(i, this.rawData[j][i], thresholds);
                    double u = getOrdinalUpperBound(i, this.rawData[j][i], thresholds);
                    if (u == Double.POSITIVE_INFINITY)
                        this.rawDataStar[j][i] = l + 1. / this.numCat[i];
                    else if (l == Double.NEGATIVE_INFINITY)
                        this.rawDataStar[j][i] = u - 1. / this.numCat[i];
                    else
                        this.rawDataStar[j][i] = r.nextDouble() / 0.5 * (u - l) + l;
                }
            } else {
                for (int j = 0; j < this.sampleSize; j++)
                    this.rawDataStar[j][i] = this.rawData[j][i];
            }
        }
                
        double impliedGInvWishartParam[][] = getImpliedGInvWishartParam(directed, this.rawData, this.rawDataStar);

        //Iteration
        for (int i = 0; i < number; i++) {
            logger.finest("- Sample #" + (i + 1));

            List<double[][]> matrices;
            matrices = sampleBidirectedMatrixGibbs(r, this.vPriorDf + this.sampleSize, impliedGInvWishartParam, bidirected);
            samplesBidirected[i] = matrices.get(0);
            double samplesBidirectedInv[][] = matrices.get(1), weight = matrices.get(2)[0][0];
            samplesDirected[i] = sampleDirectedMatrix(r, samplesBidirected[i], samplesBidirectedInv, directed,
                                                      this.rawData, this.rawDataStar);
            directed = samplesDirected[i]; bidirected = samplesBidirected[i];
            if (this.numLatents > 0) {
                //sampleLatents(r, samplesDirected[i], samplesBidirectedInv, this.rawData, this.rawDataStar);
                sampleLatentsBlock(r, samplesBidirected[i], samplesDirected[i], this.rawDataStar);
                for (int s = 0; s < this.sampleSize; s++)
                    for (int l = 0; l < this.numLatents; l++)
                        this.rawData[s][this.latentIndex[l]] = this.rawDataStar[s][this.latentIndex[l]];
                if (this.returnLatents)
                    for (int s = 0; s < this.sampleSize; s++)
                        for (int l = 0; l < this.numLatents; l++)
                            latents[i][s][l] = this.rawData[s][this.latentIndex[l]];
            }
            sampleUnderlyingVariables(r, samplesDirected[i], samplesBidirectedInv, thresholds, this.rawData, this.rawDataStar);
            sampleThresholds[i] = sampleThresholds(r, samplesDirected[i], samplesBidirected[i], samplesBidirectedInv,
                                                 thresholds, this.rawData, this.rawDataStar);

            impliedGInvWishartParam = getImpliedGInvWishartParam(samplesDirected[i], this.rawData, this.rawDataStar);
            samplesWeight[i] = weight;
            samplesBidirected[i] = packCovariance(samplesBidirected[i]);
        }

        samples.add(samplesBidirected); samples.add(samplesWeight);
        samples.add(samplesDirected); samples.add(sampleThresholds);
        if (this.returnLatents)
            samples.add(latents);

        return samples;
    }

    /**
     * Gets the matrix parameter for the G-Inverse Wishart posterior distribution using the current parameters
     * for directed edges. Unlike in McmcGaussianMixedGraph, we have to account for the fact we have two data sources:
     * the original data (with filled-in latent variables) and the "underlying latent variables" corresponding to the
     * discrete ones.
     */

    protected double[][] getImpliedGInvWishartParam(double directedParams[][], double data[][], double dataStar[][]) {
        double output[][] = new double[this.numNodes][this.numNodes];
        for (int i = 0; i < this.numNodes; i++) {
            for (int j = 0; j < this.numNodes; j++)
                output[i][j] = 0.;
        }
        for (int s = 0; s < data.length; s++) {
            for (int i = 0; i < this.numNodes; i++) {
                this.residuals[i] = dataStar[s][i];
                for (int p = 0; p < this.parents[i].length; p++)
                    this.residuals[i] -= directedParams[i][p] * data[s][this.parents[i][p]];
            }
            for (int i = 0; i < this.numNodes; i++)
                for (int j = 0; j < this.numNodes; j++)
                    output[i][j] += this.residuals[i] * this.residuals[j];
        }
        for (int i = 0; i < this.numNodes; i++)
            for (int j = 0; j < this.numNodes; j++)
                output[i][j] += this.vPrior[i][j];
        return output;
    }

    protected double[][] sampleDirectedMatrix(Random r, double errorMatrix[][],double invErrMatrix[][],
                                              double previousParams[][], double data[][], double dataStar[][]) {
        double directedMatrix[][] = new double[this.numNodes][];
        for (int i = 0; i < this.numNodes; i++) {
            directedMatrix[i] = new double[this.parents[i].length];
            System.arraycopy(previousParams[i], 0, directedMatrix[i], 0, directedMatrix[i].length);
        }
        for (int i = 0; i < this.numNodes; i++) {
            for (int pj = 0; pj < directedMatrix[i].length; pj++) {
                if (this.fixedParams[i][pj])
                    continue;
                int j = this.parents[i][pj];
                double variance = 0.;
                for (int s = 0; s < data.length; s++)
                    variance += data[s][j] * data[s][j];
                variance = 1. / (invErrMatrix[i][i] * variance + 1. / this.bPrior[i][pj][1]);
                double mean = this.bPrior[i][pj][0] / this.bPrior[i][pj][1];
                for (int s = 0; s < data.length; s++) {
                    double aux = 0.;
                    //Other parents of the same node i
                    double aux2 = dataStar[s][i];
                    for (int pOthers = 0; pOthers < directedMatrix[i].length; pOthers++) {
                        if (pOthers == pj)
                            continue;
                        aux2 -= directedMatrix[i][pOthers] * data[s][this.parents[i][pOthers]];
                    }
                    aux += invErrMatrix[i][i] * aux2;
                    //Other nodes in the same district and their parents
                    for (int pp = 0; pp < this.districtNeighbors[i].length; pp++) {
                        int p = this.districtNeighbors[i][pp];
                        aux2 = dataStar[s][p];
                        for (int pq = 0; pq < this.parents[p].length; pq++) {
                            int q = this.parents[p][pq];
                            aux2 -= directedMatrix[p][pq] * data[s][q];
                        }
                        aux2 *= invErrMatrix[i][p];
                        aux += aux2;
                    }
                    mean += data[s][j] * aux;
                }
                mean *= variance;
                directedMatrix[i][pj] = r.nextGaussian() * Math.sqrt(variance) + mean;
            }
        }

        return directedMatrix;
    }

    /**
     * The underlying variables are distributed as truncated Gaussians. Notice that each y_star has a single child y,
     * and this child does not have any other parent nor is an underlying latent variable, which simplifies
     * the posterior.
     *
     * This is virtually the same thing as sampling a latent variable, except that we should respect the
     * threshold boundaries.
     */

    protected void sampleUnderlyingVariables(Random r, double b[][], double invV[][], double thresholds[][],
                                             double data[][], double dataStar[][]) {
        for (int s = 0; s < data.length; s++)
            for (int i = 0; i < this.numNodes; i++) {
                if (!this.isDiscrete[i]) {
                    dataStar[s][i] = data[s][i];
                    continue;
                }
                double variance = 1. / invV[i][i];
                double mean = 0.;
                //Parents of the same node i
                double aux2 = b[i][this.parents[i].length - 1];
                for (int p = 0; p < this.parents[i].length - 1; p++)
                    aux2 += b[i][p] * data[s][this.parents[i][p]];
                mean += invV[i][i] * aux2;
                //Other nodes in the same district and their parents
                for (int pp = 0; pp < this.districtNeighbors[i].length; pp++) {
                    int p = this.districtNeighbors[i][pp];
                    aux2 = dataStar[s][p];
                    for (int pq = 0; pq < this.parents[p].length; pq++) {
                        int q = this.parents[p][pq];
                        aux2 -= b[p][pq] * data[s][q];
                    }
                    aux2 *= invV[i][p];
                    mean -= aux2;
                }
                mean *= variance;
                double l = getOrdinalLowerBound(i, data[s][i], thresholds), u = getOrdinalUpperBound(i, data[s][i], thresholds);
                dataStar[s][i] = RbasProbUtils.truncatedGaussianRand(mean, variance, l, u, r);
            }
    }

    /**
     * Sampling a latent variable. Unlike in the Gaussian case, implementing block sampling is not straightforward.
     * In this implementation, we sample one latent conditioning on all others. Ideally, future implementations
     * should recognize jointly Gaussian variables in order to improve the mixing of the chain.
     *
     * Unlike in the underlying variable case, a latent variable might have several children (which are not just
     * the discrete projection of itself). We should take that into account.
     */

    protected void sampleLatents(Random r, double b[][], double invV[][], double data[][], double dataStar[][]) {
        for (int s = 0; s < data.length; s++)
            for (int l = 0; l < this.numLatents; l++) {
                int i = this.latentIndex[l];

                //////////Find variance
                double invVariance = invV[i][i];
                //Find correlated children
                for (int bc1 = 0; bc1 < this.children[i].length; bc1++) {
                    int c1 = this.children[i][bc1];
                    int pidx1 = -1;
                    for (int p = 0; p < this.parents[c1].length; p++)
                        if (this.parents[c1][p] == i) {
                            pidx1 = p;
                            break;
                        }
                    if (invV[c1][i] != 0)
                        invVariance -= 2 * b[c1][pidx1] * invV[c1][i];
                    for (int bc2 = bc1; bc2 < this.children[i].length; bc2++) {
                        int c2 = this.children[i][bc2];
                        int pidx2 = -1;
                        for (int p = 0; p < this.parents[c2].length; p++)
                            if (this.parents[c2][p] == i) {
                                pidx2 = p;
                                break;
                            }
                        if (invV[c1][c2] != 0.) {
                            if (bc1 == bc2)
                                invVariance += b[c1][pidx1] * b[c2][pidx2] * invV[c1][c2];
                            else
                                invVariance += 2 * b[c1][pidx1] * b[c2][pidx2] * invV[c1][c2];
                        }
                    }
                }
                double variance = 1. / invVariance;

                //////////Find mean
                double mean = 0.;
                //Parents of the same node i
                double aux2 = 0.;
                for (int p = 0; p < this.parents[i].length; p++)
                    aux2 += b[i][p] * data[s][this.parents[i][p]];
                mean += invV[i][i] * aux2;
                //Other nodes in the same district and their parents (different from i)
                for (int pp = 0; pp < this.districtNeighbors[i].length; pp++) {
                    int p = this.districtNeighbors[i][pp];
                    aux2 = dataStar[s][p];
                    for (int pq = 0; pq < this.parents[p].length; pq++)
                        if (this.parents[p][pq] != i)
                            aux2 -= b[p][pq] * data[s][this.parents[p][pq]];
                    aux2 *= invV[i][p];
                    mean -= aux2;
                }
                //Now, iterate over children and their Markov blankets
                for (int cc = 0; cc < this.children[i].length; cc++) {
                    int c = this.children[i][cc];
                    int pidx = -1;
                    for (int pp = 0; pp < this.parents[c].length; pp++)
                        if (this.parents[c][pp] == i) {
                            pidx = pp;
                            break;
                        }
                    //other parents of the same child equation
                    aux2 = dataStar[s][c];
                    for (int pp = 0; pp < this.parents[c].length; pp++)
                        if (this.parents[c][pp] != i)
                            aux2 -= b[c][pp] * data[s][this.parents[c][pp]];
                    aux2 *= invV[c][c] * b[c][pidx];
                    mean += aux2;
                    //other members of its Markov blanket
                    for (int pp = 0; pp < this.districtNeighbors[c].length; pp++) {
                        int p = this.districtNeighbors[c][pp];
                        if (p == i)
                            continue;
                        aux2 = dataStar[s][p];
                        for (int pq = 0; pq < this.parents[p].length; pq++)
                            if (this.parents[p][pq] != i)
                                aux2 -= b[p][pq] * data[s][this.parents[p][pq]];
                        aux2 *= 0.5 * invV[c][p] * b[c][pidx];
                        mean += aux2;
                    }
                }
                mean *= variance;

                //////////Finally, sample
                data[s][i] = dataStar[s][i] = r.nextGaussian() * Math.sqrt(variance) + mean;
            }
    }

    /**
     * This need to be modified to differentiate between data and dataStar.
     */

    protected void getImpliedMoments(double v[][], double directedParams[][]) {
        for (int i = 0; i < this.numNodes; i++) {
            for (int j = 0; j < this.numNodes; j++)
                this.iMinusB[i][j] = 0.;
            this.iMinusB[i][i] = 1.;
        }
        for (int i = 0; i < this.numNodes; i++)
            for (int p = 0; p < this.parents[i].length - 1; p++)
                if (this.isLatent[this.parents[i][p]])
                    this.iMinusB[i][this.parents[i][p]] -= directedParams[i][p];
        double inv[][] = MatrixUtils.inverse(this.iMinusB);
        for (int i = 0; i < this.numNodes; i++) {
            for (int j = i; j < this.numNodes; j++) {
                this.centerMoments[i][j] = v[i][j];
                for (int p1 = 0; p1 < this.parents[i].length; p1++)
                    for (int p2 = 0; p2 < this.parents[j].length; p2++)
                        //Notice how this is different from the Gaussian case
                        if (!this.isLatent[this.parents[i][p1]] && !this.isLatent[this.parents[j][p2]])
                            this.centerMoments[i][j] += directedParams[i][p1] * directedParams[j][p2];
                this.centerMoments[j][i] = this.centerMoments[i][j];
            }
            this.eNodes[i] = 0.;
            for (int j = 0; j < this.numNodes; j++)
                this.eNodes[i] += inv[i][j] * directedParams[j][this.parents[j].length - 1];
        }
        this.eeNodes = MatrixUtils.product(inv, MatrixUtils.product(this.centerMoments, MatrixUtils.transpose(inv)));
    }

    /**
     * The thresholds are implicitly parameterized as a function of meanStar and the region vectors. See notes at the
     * top of the file. What we actually do is to sample the size of the intervals between thresholds, using a
     * variable called meanStar as a reference point to where in the real line such adjacent thresholds are located.
     */

    protected double[][] sampleThresholds(Random r, double b[][], double v[][], double invV[][], double thresholds[][],
                                          double data[][], double dataStar[][]) {
        double output[][] = new double[this.numNodes][];
        for (int i = 0; i < this.numNodes; i++)
            if (this.isDiscrete[i]) {
                output[i] = new double[this.numCat[i] - 1];
                System.arraycopy(thresholds[i], 0, output[i], 0, thresholds[i].length);
            }

        for (int i = 0; i < this.numNodes; i++) {
            for (int n = 0; n < this.numCat[i] - 1; n++) {
                if (this.fixedThreshold[i][n])
                    continue;
                double rightBoundary = Double.POSITIVE_INFINITY, leftBoundary = Double.NEGATIVE_INFINITY;
                for (int s = 0; s < data.length; s++) {
                    int p = getValueIndex(i, data[s][i]);
                    if (p == n && dataStar[s][i] > leftBoundary)
                        leftBoundary = dataStar[s][i];
                    else if (p == n + 1 && dataStar[s][i] < rightBoundary)
                        rightBoundary = dataStar[s][i];
                }
                output[i][n] = RbasProbUtils.truncatedGaussianRand(this.thresholdPrior[i][n][0], this.thresholdPrior[i][n][1],
                                                                   leftBoundary, rightBoundary, r);
            }
        }

        return output;
    }



    /******************************************************
     * Ordinal variables indexing
     */

    protected int getValueIndex(int variable, double value) {
        for (int i = 0; i < this.ordValues[variable].length; i++)
            if (this.ordValues[variable][i] == value)
                return i;
        throw new RuntimeException("Value " + value + " for ordinal variable " +
                                   this.graph.getNodes().get(variable).getName() + " not pre-specified");
    }

    protected double getOrdinalLowerBound(int variable, double value, double thresholds[][]) {
        int v = getValueIndex(variable, value);
        if (v == 0)
            return Double.NEGATIVE_INFINITY;
        return thresholds[variable][v - 1];
    }

    protected double getOrdinalUpperBound(int variable, double value, double thresholds[][]) {
        int v = getValueIndex(variable, value);
        if (v == this.numCat[variable] - 1)
            return Double.POSITIVE_INFINITY;
        return thresholds[variable][v];
    }

    /******************************************************
     * Likelihoods
     */

    /**
     * Calculates the predictive log-likelihood for the new data. The current implementation works only for
     * discrete datasets and fully observable data.
     */

    public double predictiveLoglikelihood(List inputSamples, RectangularDataSet data, int burnIn, int simulationSize, int stepSize) {
        List samples;
        if (inputSamples == null)
            samples = doSample(simulationSize);
        else
            samples = inputSamples;

        double dummyRawData[][] = data.getDoubleData().toArray();
        List<String> testVariableNames = data.getVariableNames();
        int    testSampleSize = data.getNumRows();

        Hashtable<String, Double> storedValues = new Hashtable<String, Double>();
        double pattern[] = new double[this.numObserved];
        double testLikelihood = 0.;
        for (int d = 0; d < testSampleSize; d++) {
            for (int i = 0; i < this.numObserved; i++) {
                String varName = this.graph.getNodes().get(this.observedIndex[i]).getName();
                for (int j = 0; j < this.numNodes; j++) {
                    if (testVariableNames.get(j).equals(varName)) {
                        pattern[i] = dummyRawData[d][j];
                        break;
                    }
                }
            }
            StringBuffer codedRow = new StringBuffer();
            for (int i = 0; i < this.numObserved; i++)
                codedRow.append(pattern[i] + " ");
            Double value = storedValues.get(codedRow.toString());
            if (value != null) {
                testLikelihood += value;
                continue;
            }
            double prob = evaluateProbability(samples, burnIn, pattern);
            testLikelihood += Math.log(prob);
            storedValues.put(codedRow.toString(), Math.log(prob));
            System.out.println("[" + (d + 1) + "] = " + codedRow + "= " + prob);
        }
        return testLikelihood;
    }

    /**
     * Prints every entry of the estimated mean contingency table. This of course assumes that the number of variables
     * and values is small, and that there are no continuous variables. The current implementation works only with
     * fully observable data.
     *
     * This can be highly optimized if the district size is much smaller than the number of variables. Currently,
     * this is putting all districts together.
     *
     */

    public List<double[]> impliedContingencyTable(List inputSamples, int burnIn, int simulationSize, int stepSize) {
        for (int i = 0; i < this.numObserved; i++)
            System.out.print(this.graph.getNodes().get(this.observedIndex[i]) + "\t");
        System.out.println();

        List samples;
        if (inputSamples == null)
            samples = doSample(simulationSize);
        else
            samples = inputSamples;

        List<double[]> entries = new ArrayList<double[]>();
        double entry[] = new double[this.numObserved + 1];
        for (int i = 0; i < this.numObserved; i++)
             entry[i] = 0.;

        double totalProb = 0.;

        nextEntry:
        while (true) {
            for (int i = 0; i < this.numObserved; i++)
                System.out.print((int) entry[i] + "\t");

            double pattern[] = new double[this.numObserved];
            for (int i = 0; i < this.numObserved; i++)
                pattern[i] = this.ordValues[this.observedIndex[i]][(int) entry[i]];

            double empirical = 0.;
            nextTrainingPoint:
            for (int s = 0; s < this.rawData.length; s++) {
                for (int n = 0; n < this.numObserved; n++)
                    if (this.rawData[s][this.observedIndex[n]] != pattern[n])
                        continue nextTrainingPoint;
                empirical++;
            }
            empirical /= this.rawData.length;

            entry[this.numObserved] = evaluateProbability(samples, burnIn, pattern);
            entries.add(entry);
            totalProb += entry[this.numObserved];
            System.out.println(" = " + entry[this.numObserved] + " (" + empirical + ")");

            //Move to the next combination of values
            double oldEntry[] = entry;
            entry = new double[this.numObserved + 1];
            System.arraycopy(oldEntry, 0, entry, 0, this.numObserved);
            boolean carry = false;
            for (int i = this.numObserved - 1; i >= 0; i--) {
                if (i == this.numObserved - 1)
                    entry[i]++;
                if (carry)
                    entry[i]++;
                carry = (entry[i] >= this.numCat[this.observedIndex[i]]);
                entry[i] = entry[i] % this.numCat[this.observedIndex[i]];
            }

            for (int i = 0; i < this.numObserved; i++)
                if (entry[i] != 0)
                    continue nextEntry;
            break;
        }

        System.out.println("Total probability = " + totalProb);
        return entries;
    }

    public double evaluateProbability(List samples, int burnIn, double pattern[]) {
        for (int i = 0; i < this.isDiscrete.length; i++)
            if (!this.isDiscrete[i] && !this.isLatent[i])
               throw new RuntimeException("The current implementation does not allow the computation of predictive " +
                                          "loglikelihoods for datasets with both continuous and discrete variables.");

        int simulationSize = ((double [][][]) samples.get(0)).length;

        double allBidirected[][][] = (double[][][]) samples.get(0);
        double allDirected[][][] = (double[][][]) samples.get(2);
        double allThresholds[][][] = (double[][][]) samples.get(3);

        double errSampleObs[][] = new double[this.numObserved][this.numObserved];
        double meanObs[] = new double[this.numObserved];

        double dataPoint[] = new double[this.numObserved + 1];
        System.arraycopy(pattern, 0, dataPoint, 0, this.numObserved);
        dataPoint[this.numObserved] = 1.;

        double outputProb = 0.;
        int    totalSamples = 0;

        for (int s = burnIn; s < simulationSize; s++) {
            double errSample[][] = unpackCovariance(allBidirected[s]);
            double coeffSample[][] = allDirected[s];
            double thresholds[][] = allThresholds[s];

            if (this.numLatents > 0) {
                double errL[][] = new double[this.numLatents][this.numLatents];
                for (int l1 = 0; l1 < this.numLatents; l1++)
                    for (int l2 = l1; l2 < this.numLatents; l2++)
                        errL[l1][l2] = errL[l2][l1] = errSample[this.latentIndex[l1]][this.latentIndex[l2]];
                double iMinusBL[][] = new double[this.numLatents][this.numLatents];
                for (int l = 0; l < this.numLatents; l++) {
                    int li = this.latentIndex[l];
                    iMinusBL[l][l] = 1.;
                    for (int lp = 0; lp < this.parents[li].length - 1; lp++) {
                        int p = this.parents[li][lp];
                        if (this.isLatent[p])
                            iMinusBL[l][this.latentIndexReverse[p]] = -coeffSample[li][lp];
                    }
                }
                double invIMinusBL[][] = MatrixUtils.inverse(iMinusBL);
                double covL[][] = MatrixUtils.product(invIMinusBL,
                                  MatrixUtils.product(errL, MatrixUtils.transpose(invIMinusBL)));
                double meanAux[] = new double[this.numLatents];
                for (int l = 0; l < this.numLatents; l++) {
                    int li = this.latentIndex[l];
                    meanAux[l] = coeffSample[li][this.parents[li].length - 1];
                    for (int lp = 0; lp < this.parents[li].length - 1; lp++)
                        if (!this.isLatent[this.parents[li][lp]])
                            meanAux[l] += coeffSample[li][lp] *
                                          dataPoint[this.observedIndexReverse[this.parents[li][lp]]];
                }
                double meanL[] = MatrixUtils.product(invIMinusBL, meanAux);
                double bL[][] = new double[this.numObserved][this.numLatents];
                for (int v = 0; v < this.numObserved; v++) {
                    int lv = this.observedIndex[v];
                    for (int p = 0; p < this.parents[lv].length - 1; p++) {
                        if (!this.isLatent[this.parents[lv][p]])
                            continue;
                        bL[v][this.latentIndexReverse[this.parents[lv][p]]] = coeffSample[lv][p];
                    }
                }
                double eeL[][] = new double[this.numLatents][this.numLatents];
                for (int l1 = 0; l1 < this.numLatents; l1++)
                    for (int l2 = 0; l2 < this.numLatents; l2++)
                        eeL[l1][l2] = covL[l1][l2] + meanL[l1] * meanL[l2];
                double covAux[][] = MatrixUtils.product(bL, MatrixUtils.product(eeL, MatrixUtils.transpose(bL)));
                for (int i = 0; i < this.numObserved; i++) {
                    int v = this.observedIndex[i];
                    meanObs[i] = coeffSample[v][this.parents[v].length - 1];
                    for (int p = 0; p < this.parents[v].length - 1; p++) {
                        if (!this.isLatent[this.parents[v][p]])
                            meanObs[i] += coeffSample[v][p] *
                                          dataPoint[this.observedIndexReverse[this.parents[v][p]]];
                        else
                            meanObs[i] += coeffSample[v][p] *
                                          meanL[this.latentIndexReverse[this.parents[v][p]]];
                    }
                    for (int j = 0; j < this.numObserved; j++)
                        errSampleObs[i][j] = errSample[this.observedIndex[i]][this.observedIndex[j]] +
                                             covAux[i][j];
                    //TODO: THIS ASSUMES THAT LATENTS AND ERROR TERMS INDEPENDENT
                }
            } else {
                for (int i = 0; i < this.numObserved; i++) {
                    meanObs[i] = 0.;
                    for (int p = 0; p < this.parents[i].length; p++)
                        meanObs[i] += coeffSample[i][p] * dataPoint[this.parents[i][p]];
                    for (int j = 0; j < this.numObserved; j++)
                        errSampleObs[i][j] = errSample[this.observedIndex[i]][this.observedIndex[j]];
                }
            }

            double corr[][] = new double[this.numObserved][this.numObserved];
            double tUpper[] = new double[this.numObserved];
            double tLower[] = new double[this.numObserved];
            for (int i = 0; i < this.numObserved; i++) {
                for (int j = 0; j < this.numObserved; j++)
                    corr[i][j] = errSampleObs[i][j] / Math.sqrt(errSampleObs[i][i] * errSampleObs[j][j]);
                tUpper[i] = (getOrdinalUpperBound(this.observedIndex[i], dataPoint[i], thresholds) - meanObs[i])
                          / Math.sqrt(errSampleObs[i][i]);
                tLower[i] = (getOrdinalLowerBound(this.observedIndex[i], dataPoint[i], thresholds) - meanObs[i])
                          / Math.sqrt(errSampleObs[i][i]);
            }
            double currentProb = RbasProbUtils.multinormalProb(tLower, tUpper, corr);
            outputProb += currentProb;
            totalSamples++;
        }

        outputProb /= totalSamples;

        return outputProb;
    }


    /*************
     * Other methods
     */

    public double[][] getOrdValues() {
        return this.ordValues;
    }

    public int[] getNumCat() {
        return this.numCat;
    }
}

