package rbas.util;

import edu.cmu.tetrad.graph.*;

import java.util.ArrayList;
import java.util.List;
import java.util.HashSet;

/**
 * RbasGraphUtils
 *
 * Series of useful (and not so useful) methods concerning graphical properties and operations on graphs,
 * particularly DMGs. This is done in order to complement Tetrad's GraphUtils class.
 *
 * @author Ricardo Silva
 * @version $Revision: 1.0 $ $Date: 2006/06/14$
 */


public class RbasGraphUtils {

    /**
     * Get Markov blankets of nodes in a dmg. The result is returned in a two-dimensional array ordered according
     * to the order of the nodes in the graph. Each row i contains the node positions of the items in the Markov
     * blanket of node i, also according to the order in the graph.getNodes() list.
     */

    public static int[][] getMarkovBlanketsDmg(Graph graph) {
        int numNodes = graph.getNumNodes();
        int blanket[][] = new int[numNodes][];
        for (int i = 0; i < numNodes; i++) {
            Node node = graph.getNodes().get(i);
            List<Node> blanketLocal = new ArrayList<Node>();
            blanketLocal.addAll(graph.getParents(node));
            blanketLocal.addAll(graph.getChildren(node));
            for (Node child : graph.getChildren(node))
                RbasGraphUtils.expandNodesIntoTarget(graph, node, child, blanketLocal);
            blanket[i] = new int[blanketLocal.size()];
            int bc = 0;
            for (Node blanketItem : blanketLocal) {
                for (int n = 0; n < graph.getNodes().size(); n++) {
                    Node nodeBlanket = graph.getNodes().get(n);
                    if (nodeBlanket.getName().equals(blanketItem.getName())) {
                        blanket[i][bc++] = n;
                        break;
                    }
                }
            }
        }
        return blanket;
    }

    /**
     * Search for nodes that are adjacent to target by an edge into target (i.e., either parents or
     * spouses of target), except the source node that triggered this search. Do it recursively.
     */

    private static void expandNodesIntoTarget(Graph graph, Node source, Node target, List blanket) {
        List<Node> intoNeighbors = new ArrayList<Node>();
        for (Node neighbor : graph.getAdjacentNodes(target)) {
            if (!blanket.contains(neighbor) && !graph.isChildOf(neighbor, target) && neighbor != source) {
                intoNeighbors.add(neighbor);
                blanket.add(neighbor);
            }
        }
        for (Node next : intoNeighbors)
            if (!graph.isParentOf(next, target))
                RbasGraphUtils.expandNodesIntoTarget(graph, source, next, blanket);
    }

    /**
     * Get the graph composed of the nodes and bi-directed edges only.
     */

    public static Graph getBidirectedSubgraph(Graph g) {
        Graph graph = new EdgeListGraph(g);
        List<Edge> edgesToRemove = new ArrayList<Edge>();
        for (Edge edge : g.getEdges())
            if (edge.getEndpoint1() != Endpoint.ARROW || edge.getEndpoint2() != Endpoint.ARROW)
                edgesToRemove.add(edge);
        graph.removeEdges(edgesToRemove);
        return graph;
    }

    /**
     * Returns true if graph is bidirected
     */

    public static boolean isBidirected(Graph g) {
        for (Edge e : g.getEdges()) {
            if (e.getEndpoint1() != Endpoint.ARROW || e.getEndpoint2() != Endpoint.ARROW)
                return false;
        }
        return true;
    }

    public static void printEdges(Graph graph) {
        for (Edge edge : graph.getEdges())
             System.out.println(" -- " + edge.toString());
    }

    public static int[][] getParentsMatrix(Graph graph) {
        int parents[][] = new int[graph.getNumNodes()][];
        for (int i = 0; i < graph.getNumNodes(); i++) {
            Node node = (Node) graph.getNodes().get(i);
            List p = graph.getParents(node);
            parents[i] = new int[p.size()];
            for (int pc = 0; pc < p.size(); pc++) {
                Node parent = (Node) p.get(pc);
                for (int j = 0; j < graph.getNumNodes(); j++) {
                    Node cparent = (Node) graph.getNodes().get(j);
                    if (cparent == parent) {
                        parents[i][pc] = j;
                        break;
                    }
                }
            }
        }
        return parents;
    }

    public static int[][] getChildrenMatrix(Graph graph) {
        int children[][] = new int[graph.getNumNodes()][];
        for (int i = 0; i < graph.getNumNodes(); i++) {
            Node node = (Node) graph.getNodes().get(i);
            List c = graph.getChildren(node);
            children[i] = new int[c.size()];
            for (int pc = 0; pc < c.size(); pc++) {
                Node child = (Node) c.get(pc);
                for (int j = 0; j < graph.getNumNodes(); j++) {
                    Node cchild = (Node) graph.getNodes().get(j);
                    if (cchild == child) {
                        children[i][pc] = j;
                        break;
                    }
                }
            }
        }
        return children;
    }

    /****************************************************************************************************************
     * GRAPH FACTORIZATION: we will find all districts in the graph, and reorder the nodes within the
     * graph data structure to correspond to the factors. This can achieve a great improvement in computational
     * performance if the model is large but with small districts. THIS IS FOR ACYCLIC GRAPHS ONLY.
     */

    /**
     * Methods for finding the districts of the graph (Richardson, 2003). Important step on factorizing the
     * error covariance matrix (i.e., if it is a diagonal one, then each node is a district in itself and no
     * special decomposition will be necessary when sampling.)
     */

    public    static List factorizeGraph(Graph graph, boolean verbose) {
        List output = new ArrayList();
        List<List<Node>> districtList = new ArrayList<List<Node>>();
        HashSet<Node>    included     = new HashSet<Node>();

        for (Node target : graph.getNodes()) {
            if (included.contains(target))
                continue;
            List<Node> district = new ArrayList<Node>();
            district.add(target);
            expandDistrict(target, district, graph);
            for (Node member : district)
                included.add(member);
            districtList.add(district);
        }

        int numDistricts = districtList.size();
        int numNodes = graph.getNumNodes();

        int districtNeighbors[][] = new int[numNodes][];
        int districts[][] = new int[numDistricts][];

        for (List<Node> district : districtList) {
            //Order each district in a way that each node comes before its parents in the district
            HashSet<Node> elements = new HashSet<Node>();
            for (Node node : district)
                elements.add(node);
            Node          items[] = new Node[elements.size()];
            HashSet<Node> parents[] = new HashSet[elements.size()];
            int           countNode = 0;
            for (Node node : district)  {
                items[countNode] = node;
                parents[countNode] = new HashSet<Node>();
                List<Node> p = graph.getParents(node);
                for (Node parent : p)
                    if (elements.contains(parent))
                        parents[countNode].add(parent);
                countNode++;
            }
            district.clear();
            boolean eliminated[] = new boolean[elements.size()];
            int     numEliminated = 0;
            while (numEliminated < elements.size()) {
                for (int i = 0; i < elements.size(); i++) {
                    if (!eliminated[i] && parents[i].size() == 0) {
                        district.add(items[i]);
                        for (int j = 0; j < elements.size(); j++)
                            if (parents[j].contains(items[i]))
                                parents[j].remove(items[i]);
                        eliminated[i] = true;
                        numEliminated++;
                    }
                }
            }
        }

        //Tries to reorder districts, in a way that a district X will not preceed district Y if Y contains some
        //parent of some element of X. This is not possible in general, even if the graph is acyclic. But if true,
        //is useful both for display purposes and to order the coefficient matrix as a lower triangular matrix
        //(easy to invert)
        Graph districtGraph = new EdgeListGraph();
        for (int i = 0; i < districtList.size(); i++)
            districtGraph.addNode(new GraphNode("0" + i));
        for (int i = 0; i < districtList.size(); i++)
            nextPair:
            for (int j = 0; j < districtList.size(); j++) {
                if (i == j)
                    continue;
                for (Node node1 : districtList.get(i)) {
                    for (Node node2 : districtList.get(j))
                        if (graph.isParentOf(node1, node2)) {
                            districtGraph.addDirectedEdge(districtGraph.getNode("0" + i), districtGraph.getNode("0" + j));
                            continue nextPair;
                        }
                }
            }
        if (!districtGraph.existsDirectedCycle()) {
            List<List<Node>> reorderedDistrict = new ArrayList<List<Node>>();
            while (districtGraph.getNumNodes() > 0) {
                Node chosen = null;
                for (Node d : districtGraph.getNodes())
                    if (districtGraph.getParents(d).size() == 0) {
                        chosen = d;
                        break;
                    }
                int districtNumber = Integer.parseInt(chosen.getName());
                reorderedDistrict.add(districtList.get(districtNumber));
                districtGraph.removeNode(chosen);
            }
            districtList.clear(); districtList.addAll(reorderedDistrict);
        }


        int distCount = 0, nodeCount = 0;
        if (verbose)
            System.out.println("-- DISTRICTS: ");
        for (List<Node> district : districtList) {
            //Now that everything is ordered, number the nodes
            districts[distCount] = new int[district.size()];
            int distInnerCount = 0;
            for (Node node : district) {
                if (verbose)
                    System.out.print(node.toString() + " ");
                districts[distCount][distInnerCount++] = nodeCount++;
            }
            if (verbose)
                System.out.println();
            distCount++;
        }

        for (int i = 0; i < districts.length; i++) {
            for (int j = 0; j < districts[i].length; j++) {
                districtNeighbors[districts[i][j]] = new int[districts[i].length - 1];
                int count = 0;
                for (int k = 0; k < districts[i].length; k++) {
                    if (j != k)
                       districtNeighbors[districts[i][j]][count++] = districts[i][k];
                }
            }
        }

        Graph outputGraph = reorderGraph(graph, districtList);
        output.add(outputGraph); output.add(districts); output.add(districtNeighbors);

        return output;
    }

    protected static void expandDistrict(Node target, List district, Graph graph) {
        List<Node> preSpouses = new ArrayList<Node>();
        for (Node neighbor : graph.getAdjacentNodes(target)) {
            if (!district.contains(neighbor) &&
                    graph.getEndpoint(neighbor, target) == Endpoint.ARROW &&
                    graph.getEndpoint(target, neighbor) == Endpoint.ARROW) {
                preSpouses.add(neighbor);
                district.add(neighbor);
            }
        }
        for (Node next : preSpouses)
            expandDistrict(next, district, graph);
    }

    protected static Graph reorderGraph(Graph graph, List<List<Node>> districtList) {
        Graph output = new EdgeListGraph();
        for (List<Node> district : districtList)
            for (Node node : district)
                output.addNode(node);
        for (Edge next : graph.getEdges())
             output.addEdge(next);
        return output;
    }


    /**
     * Return nodes in dag according to the partial order where no node antecedes one of its
     * ancestors.
     */

    public static List<Node> getPartialOrder(Graph dag) {
        Graph dagCopy = new EdgeListGraph(dag);
        List<Node> output = new ArrayList<Node>();
        while (dagCopy.getNumNodes() > 0) {
            List<Node> exogenous = new ArrayList<Node>();
            for (Node node : dagCopy.getNodes())
                if (dagCopy.getParents(node).size() == 0)
                    exogenous.add(node);
            output.addAll(exogenous);
            dagCopy.removeNodes(exogenous);
        }
        return output;
    }

    /**
     * Returns the number of a node within a graph.
     */

    public static int getNodeNumber(Graph graph, Node toFind) {
        int i = 0;
        for (Node node : graph.getNodes()) {
            if (node == toFind)
                return i;
            i++;
        }
        return -1;
    }
}
